# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ScheduleConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cron_expression': 'str',
        'timezone': 'str'
    }

    attribute_map = {
        'cron_expression': 'cron_expression',
        'timezone': 'timezone'
    }

    def __init__(self, cron_expression=None, timezone='Universal', local_vars_configuration=None):  # noqa: E501
        """ScheduleConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cron_expression = None
        self._timezone = None
        self.discriminator = None

        self.cron_expression = cron_expression
        if timezone is not None:
            self.timezone = timezone

    @property
    def cron_expression(self):
        """Gets the cron_expression of this ScheduleConfig.  # noqa: E501

        A cron expression to define the frequency at which to run this cron job, for example '0 0 * * *' is a cron expression that means 'run at midnight'. Visit crontab.guru to construct a precise cron_expression.  # noqa: E501

        :return: The cron_expression of this ScheduleConfig.  # noqa: E501
        :rtype: str
        """
        return self._cron_expression

    @cron_expression.setter
    def cron_expression(self, cron_expression):
        """Sets the cron_expression of this ScheduleConfig.

        A cron expression to define the frequency at which to run this cron job, for example '0 0 * * *' is a cron expression that means 'run at midnight'. Visit crontab.guru to construct a precise cron_expression.  # noqa: E501

        :param cron_expression: The cron_expression of this ScheduleConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cron_expression is None:  # noqa: E501
            raise ValueError("Invalid value for `cron_expression`, must not be `None`")  # noqa: E501

        self._cron_expression = cron_expression

    @property
    def timezone(self):
        """Gets the timezone of this ScheduleConfig.  # noqa: E501

        The timezone in which to interpret the cron_expression. Default is Universal time (UTC).  # noqa: E501

        :return: The timezone of this ScheduleConfig.  # noqa: E501
        :rtype: str
        """
        return self._timezone

    @timezone.setter
    def timezone(self, timezone):
        """Sets the timezone of this ScheduleConfig.

        The timezone in which to interpret the cron_expression. Default is Universal time (UTC).  # noqa: E501

        :param timezone: The timezone of this ScheduleConfig.  # noqa: E501
        :type: str
        """

        self._timezone = timezone

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ScheduleConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ScheduleConfig):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass
import os
import re
from typing import Optional

import pytest
import tzlocal

from anyscale.job.models import JobConfig
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus


@dataclass
class ScheduleConfigFile:
    name: str
    expected_config: Optional[ScheduleConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_files/schedule_config_files", self.name
        )


TEST_CONFIG_FILES = [
    ScheduleConfigFile(
        "empty.yaml",
        expected_error=re.escape(
            "__init__() missing 2 required positional arguments: 'job_config' and 'cron_expression'"
        ),
    ),
    ScheduleConfigFile(
        "minimal.yaml",
        expected_config=ScheduleConfig(
            cron_expression="0 0 * * * *",
            job_config=JobConfig(entrypoint="python test.py"),
        ),
    ),
    ScheduleConfigFile(
        "full.yaml",
        expected_config=ScheduleConfig(
            cron_expression="0 0 * * * *",
            timezone="local",
            job_config=JobConfig(
                name="test-name-from-file",
                image_uri="docker.io/library/test:latest",
                compute_config="test-compute-config",
                working_dir="test-working-dir",
                excludes=["test"],
                requirements=["pip-install-test"],
                entrypoint="python test.py",
                max_retries=5,
            ),
        ),
    ),
    ScheduleConfigFile(
        "unrecognized_option.yaml",
        expected_error=re.escape(
            "__init__() got an unexpected keyword argument 'bad_option'"
        ),
    ),
]


class TestScheduleConfig:
    def test_invalid_config(self):
        with pytest.raises(TypeError, match="'job_config' must be a JobConfig"):
            ScheduleConfig(job_config=None, cron_expression="0 0 * * * *")

        with pytest.raises(TypeError, match="'job_config' must be a JobConfig"):
            ScheduleConfig(job_config="job_config", cron_expression="0 0 * * * *")

    def test_invalid_cron_expression(self):
        job_config = JobConfig(entrypoint="python main.py")

        with pytest.raises(TypeError, match="'cron_expression' must be a string"):
            ScheduleConfig(job_config=job_config, cron_expression=None)

        with pytest.raises(ValueError, match="'cron_expression' cannot be empty"):
            ScheduleConfig(job_config=job_config, cron_expression="")

    def test_invalid_timezone(self):
        job_config = JobConfig(entrypoint="python main.py")
        cron_exp = "0 0 * * * *"
        with pytest.raises(TypeError, match="'timezone' must be a string"):
            ScheduleConfig(
                job_config=job_config, cron_expression=cron_exp, timezone=123,
            )

        with pytest.raises(ValueError, match="'timezone' cannot be empty"):
            ScheduleConfig(job_config=job_config, cron_expression=cron_exp, timezone="")

    def test_basic(self):
        job_config = JobConfig(entrypoint="python main.py")
        cron_exp = "0 0 * * * *"

        conf = ScheduleConfig(job_config=job_config, cron_expression=cron_exp)
        assert conf.timezone == "UTC"
        assert conf.cron_expression == cron_exp

        conf = ScheduleConfig(
            job_config=job_config, cron_expression=cron_exp, timezone="local"
        )
        assert conf.timezone == tzlocal.get_localzone_name()
        assert conf.cron_expression == cron_exp

        tz_name = "America/Los_Angeles"
        conf = ScheduleConfig(
            job_config=job_config, cron_expression=cron_exp, timezone=tz_name
        )
        assert conf.timezone == tz_name
        assert conf.cron_expression == cron_exp

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_yaml(self, config_file):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                ScheduleConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == ScheduleConfig.from_yaml(
            config_file.get_path()
        )


class TestScheduleStatus:
    @pytest.mark.parametrize(
        "state", [ScheduleState.ENABLED, ScheduleState.DISABLED,],
    )  # type: ignore
    def test_version_states(self, state: ScheduleState):
        # id, name, state, config
        assert (
            ScheduleStatus(
                id="test-schedule-id",
                name="test-schedule-name",
                state=state,
                config=ScheduleConfig(
                    cron_expression="0 0 * * * *",
                    job_config=JobConfig(entrypoint="python test.py"),
                ),
            ).state
            == state
        )

    def test_unknown_states(self):
        with pytest.raises(
            ValueError, match="'NOT_REAL_STATE' is not a valid ScheduleState"
        ):
            ScheduleStatus(
                id="test-schedule-id",
                name="test-schedule-name",
                state="NOT_REAL_STATE",
                config=ScheduleConfig(
                    cron_expression="0 0 * * * *",
                    job_config=JobConfig(entrypoint="python test.py"),
                ),
            )



from dataclasses import dataclass, field
from typing import Optional, Union

from anyscale._private.models.model_base import ModelBase, ModelEnum


class ImageBuildStatus(ModelEnum):
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"

    def __str__(self):
        return self.name

    __docstrings__ = {
        IN_PROGRESS: "The image build is in progress.",
        SUCCEEDED: "The image build succeeded.",
        FAILED: "The image build failed.",
        UNKNOWN: "The CLI/SDK received an unexpected state from the API server. In most cases, this means you need to update the CLI.",
    }


@dataclass(frozen=True)
class ImageBuild(ModelBase):
    __doc_py_example__ = """\
import anyscale
from anyscale.models import ImageBuild, ImageBuildStatus

image_build: ImageBuild = anyscale.image.get("image-name")
"""
    __doc_cli_example__ = """\
$ anyscale image get -n my-image
uri: anyscale/image/my-image:2
status: SUCCEEDED
"""

    uri: str = field(metadata={"docstring": "The URI of the image for the build."},)

    def _validate_uri(self, uri: str):
        if uri is None or not isinstance(uri, str):
            raise ValueError("The URI of the image must be a string.")

    status: Union[str, ImageBuildStatus] = field(
        metadata={"docstring": "The status of the image build."},
    )

    def _validate_status(self, status: Union[str, ImageBuildStatus]):
        return ImageBuildStatus.validate(status)

    ray_version: Optional[str] = field(
        metadata={"docstring": "The Ray version used for the image build."},
    )

    def _validate_ray_version(self, ray_version: Optional[str]):
        if ray_version is not None and not isinstance(ray_version, str):
            raise ValueError("The Ray version must be a string.")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CloudResourceGCP(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'gcp_vpc_id': 'str',
        'gcp_subnet_ids': 'list[str]',
        'gcp_cluster_node_service_account_email': 'str',
        'gcp_anyscale_iam_service_account_email': 'str',
        'gcp_filestore_config': 'GCPFileStoreConfig',
        'gcp_firewall_policy_ids': 'list[str]',
        'gcp_cloud_storage_bucket_id': 'str',
        'gcp_deployment_manager_id': 'str',
        'memorystore_instance_config': 'GCPMemorystoreInstanceConfig',
        'compute_stack': 'ComputeStack',
        'id': 'str',
        'cloud_id': 'str'
    }

    attribute_map = {
        'gcp_vpc_id': 'gcp_vpc_id',
        'gcp_subnet_ids': 'gcp_subnet_ids',
        'gcp_cluster_node_service_account_email': 'gcp_cluster_node_service_account_email',
        'gcp_anyscale_iam_service_account_email': 'gcp_anyscale_iam_service_account_email',
        'gcp_filestore_config': 'gcp_filestore_config',
        'gcp_firewall_policy_ids': 'gcp_firewall_policy_ids',
        'gcp_cloud_storage_bucket_id': 'gcp_cloud_storage_bucket_id',
        'gcp_deployment_manager_id': 'gcp_deployment_manager_id',
        'memorystore_instance_config': 'memorystore_instance_config',
        'compute_stack': 'compute_stack',
        'id': 'id',
        'cloud_id': 'cloud_id'
    }

    def __init__(self, gcp_vpc_id=None, gcp_subnet_ids=None, gcp_cluster_node_service_account_email=None, gcp_anyscale_iam_service_account_email=None, gcp_filestore_config=None, gcp_firewall_policy_ids=None, gcp_cloud_storage_bucket_id=None, gcp_deployment_manager_id=None, memorystore_instance_config=None, compute_stack=None, id=None, cloud_id=None, local_vars_configuration=None):  # noqa: E501
        """CloudResourceGCP - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._gcp_vpc_id = None
        self._gcp_subnet_ids = None
        self._gcp_cluster_node_service_account_email = None
        self._gcp_anyscale_iam_service_account_email = None
        self._gcp_filestore_config = None
        self._gcp_firewall_policy_ids = None
        self._gcp_cloud_storage_bucket_id = None
        self._gcp_deployment_manager_id = None
        self._memorystore_instance_config = None
        self._compute_stack = None
        self._id = None
        self._cloud_id = None
        self.discriminator = None

        self.gcp_vpc_id = gcp_vpc_id
        self.gcp_subnet_ids = gcp_subnet_ids
        self.gcp_cluster_node_service_account_email = gcp_cluster_node_service_account_email
        self.gcp_anyscale_iam_service_account_email = gcp_anyscale_iam_service_account_email
        self.gcp_filestore_config = gcp_filestore_config
        self.gcp_firewall_policy_ids = gcp_firewall_policy_ids
        self.gcp_cloud_storage_bucket_id = gcp_cloud_storage_bucket_id
        if gcp_deployment_manager_id is not None:
            self.gcp_deployment_manager_id = gcp_deployment_manager_id
        if memorystore_instance_config is not None:
            self.memorystore_instance_config = memorystore_instance_config
        if compute_stack is not None:
            self.compute_stack = compute_stack
        self.id = id
        self.cloud_id = cloud_id

    @property
    def gcp_vpc_id(self):
        """Gets the gcp_vpc_id of this CloudResourceGCP.  # noqa: E501

        The GCP vpc id of this cloud.  # noqa: E501

        :return: The gcp_vpc_id of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._gcp_vpc_id

    @gcp_vpc_id.setter
    def gcp_vpc_id(self, gcp_vpc_id):
        """Sets the gcp_vpc_id of this CloudResourceGCP.

        The GCP vpc id of this cloud.  # noqa: E501

        :param gcp_vpc_id: The gcp_vpc_id of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and gcp_vpc_id is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_vpc_id`, must not be `None`")  # noqa: E501

        self._gcp_vpc_id = gcp_vpc_id

    @property
    def gcp_subnet_ids(self):
        """Gets the gcp_subnet_ids of this CloudResourceGCP.  # noqa: E501

        A list of GCP subnet IDs of this cloud.  # noqa: E501

        :return: The gcp_subnet_ids of this CloudResourceGCP.  # noqa: E501
        :rtype: list[str]
        """
        return self._gcp_subnet_ids

    @gcp_subnet_ids.setter
    def gcp_subnet_ids(self, gcp_subnet_ids):
        """Sets the gcp_subnet_ids of this CloudResourceGCP.

        A list of GCP subnet IDs of this cloud.  # noqa: E501

        :param gcp_subnet_ids: The gcp_subnet_ids of this CloudResourceGCP.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and gcp_subnet_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_subnet_ids`, must not be `None`")  # noqa: E501

        self._gcp_subnet_ids = gcp_subnet_ids

    @property
    def gcp_cluster_node_service_account_email(self):
        """Gets the gcp_cluster_node_service_account_email of this CloudResourceGCP.  # noqa: E501

        The GCP service account email of this cloud.  # noqa: E501

        :return: The gcp_cluster_node_service_account_email of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._gcp_cluster_node_service_account_email

    @gcp_cluster_node_service_account_email.setter
    def gcp_cluster_node_service_account_email(self, gcp_cluster_node_service_account_email):
        """Sets the gcp_cluster_node_service_account_email of this CloudResourceGCP.

        The GCP service account email of this cloud.  # noqa: E501

        :param gcp_cluster_node_service_account_email: The gcp_cluster_node_service_account_email of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and gcp_cluster_node_service_account_email is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_cluster_node_service_account_email`, must not be `None`")  # noqa: E501

        self._gcp_cluster_node_service_account_email = gcp_cluster_node_service_account_email

    @property
    def gcp_anyscale_iam_service_account_email(self):
        """Gets the gcp_anyscale_iam_service_account_email of this CloudResourceGCP.  # noqa: E501

        The GCP service account email of this cloud.  # noqa: E501

        :return: The gcp_anyscale_iam_service_account_email of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._gcp_anyscale_iam_service_account_email

    @gcp_anyscale_iam_service_account_email.setter
    def gcp_anyscale_iam_service_account_email(self, gcp_anyscale_iam_service_account_email):
        """Sets the gcp_anyscale_iam_service_account_email of this CloudResourceGCP.

        The GCP service account email of this cloud.  # noqa: E501

        :param gcp_anyscale_iam_service_account_email: The gcp_anyscale_iam_service_account_email of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and gcp_anyscale_iam_service_account_email is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_anyscale_iam_service_account_email`, must not be `None`")  # noqa: E501

        self._gcp_anyscale_iam_service_account_email = gcp_anyscale_iam_service_account_email

    @property
    def gcp_filestore_config(self):
        """Gets the gcp_filestore_config of this CloudResourceGCP.  # noqa: E501

        The GCP filestore config of this cloud.  # noqa: E501

        :return: The gcp_filestore_config of this CloudResourceGCP.  # noqa: E501
        :rtype: GCPFileStoreConfig
        """
        return self._gcp_filestore_config

    @gcp_filestore_config.setter
    def gcp_filestore_config(self, gcp_filestore_config):
        """Sets the gcp_filestore_config of this CloudResourceGCP.

        The GCP filestore config of this cloud.  # noqa: E501

        :param gcp_filestore_config: The gcp_filestore_config of this CloudResourceGCP.  # noqa: E501
        :type: GCPFileStoreConfig
        """
        if self.local_vars_configuration.client_side_validation and gcp_filestore_config is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_filestore_config`, must not be `None`")  # noqa: E501

        self._gcp_filestore_config = gcp_filestore_config

    @property
    def gcp_firewall_policy_ids(self):
        """Gets the gcp_firewall_policy_ids of this CloudResourceGCP.  # noqa: E501

        A list of GCP firewall policy IDs of this cloud.  # noqa: E501

        :return: The gcp_firewall_policy_ids of this CloudResourceGCP.  # noqa: E501
        :rtype: list[str]
        """
        return self._gcp_firewall_policy_ids

    @gcp_firewall_policy_ids.setter
    def gcp_firewall_policy_ids(self, gcp_firewall_policy_ids):
        """Sets the gcp_firewall_policy_ids of this CloudResourceGCP.

        A list of GCP firewall policy IDs of this cloud.  # noqa: E501

        :param gcp_firewall_policy_ids: The gcp_firewall_policy_ids of this CloudResourceGCP.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and gcp_firewall_policy_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_firewall_policy_ids`, must not be `None`")  # noqa: E501

        self._gcp_firewall_policy_ids = gcp_firewall_policy_ids

    @property
    def gcp_cloud_storage_bucket_id(self):
        """Gets the gcp_cloud_storage_bucket_id of this CloudResourceGCP.  # noqa: E501

        The GCP cloud storage bucket id of this cloud.  # noqa: E501

        :return: The gcp_cloud_storage_bucket_id of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._gcp_cloud_storage_bucket_id

    @gcp_cloud_storage_bucket_id.setter
    def gcp_cloud_storage_bucket_id(self, gcp_cloud_storage_bucket_id):
        """Sets the gcp_cloud_storage_bucket_id of this CloudResourceGCP.

        The GCP cloud storage bucket id of this cloud.  # noqa: E501

        :param gcp_cloud_storage_bucket_id: The gcp_cloud_storage_bucket_id of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and gcp_cloud_storage_bucket_id is None:  # noqa: E501
            raise ValueError("Invalid value for `gcp_cloud_storage_bucket_id`, must not be `None`")  # noqa: E501

        self._gcp_cloud_storage_bucket_id = gcp_cloud_storage_bucket_id

    @property
    def gcp_deployment_manager_id(self):
        """Gets the gcp_deployment_manager_id of this CloudResourceGCP.  # noqa: E501

        The GCP deployment manager id of this cloud.  # noqa: E501

        :return: The gcp_deployment_manager_id of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._gcp_deployment_manager_id

    @gcp_deployment_manager_id.setter
    def gcp_deployment_manager_id(self, gcp_deployment_manager_id):
        """Sets the gcp_deployment_manager_id of this CloudResourceGCP.

        The GCP deployment manager id of this cloud.  # noqa: E501

        :param gcp_deployment_manager_id: The gcp_deployment_manager_id of this CloudResourceGCP.  # noqa: E501
        :type: str
        """

        self._gcp_deployment_manager_id = gcp_deployment_manager_id

    @property
    def memorystore_instance_config(self):
        """Gets the memorystore_instance_config of this CloudResourceGCP.  # noqa: E501

        The GCP memorystore instance config of this cloud.  # noqa: E501

        :return: The memorystore_instance_config of this CloudResourceGCP.  # noqa: E501
        :rtype: GCPMemorystoreInstanceConfig
        """
        return self._memorystore_instance_config

    @memorystore_instance_config.setter
    def memorystore_instance_config(self, memorystore_instance_config):
        """Sets the memorystore_instance_config of this CloudResourceGCP.

        The GCP memorystore instance config of this cloud.  # noqa: E501

        :param memorystore_instance_config: The memorystore_instance_config of this CloudResourceGCP.  # noqa: E501
        :type: GCPMemorystoreInstanceConfig
        """

        self._memorystore_instance_config = memorystore_instance_config

    @property
    def compute_stack(self):
        """Gets the compute_stack of this CloudResourceGCP.  # noqa: E501

        The compute stack associated with this cloud resource.  # noqa: E501

        :return: The compute_stack of this CloudResourceGCP.  # noqa: E501
        :rtype: ComputeStack
        """
        return self._compute_stack

    @compute_stack.setter
    def compute_stack(self, compute_stack):
        """Sets the compute_stack of this CloudResourceGCP.

        The compute stack associated with this cloud resource.  # noqa: E501

        :param compute_stack: The compute_stack of this CloudResourceGCP.  # noqa: E501
        :type: ComputeStack
        """

        self._compute_stack = compute_stack

    @property
    def id(self):
        """Gets the id of this CloudResourceGCP.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this CloudResourceGCP.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CloudResourceGCP.  # noqa: E501

        The ID of the cloud that this cloud resource belongs to.  # noqa: E501

        :return: The cloud_id of this CloudResourceGCP.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CloudResourceGCP.

        The ID of the cloud that this cloud resource belongs to.  # noqa: E501

        :param cloud_id: The cloud_id of this CloudResourceGCP.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CloudResourceGCP):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CloudResourceGCP):
            return True

        return self.to_dict() != other.to_dict()



from collections import defaultdict
import os
from typing import Dict, Generator, Optional
import uuid

import click
from click.testing import CliRunner
import pytest

from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.schedule_commands import apply, pause, resume, status, trigger
from anyscale.job.models import JobConfig
from anyscale.schedule.commands import _SCHEDULE_SDK_SINGLETON_KEY
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath)


EMPTY_CONFIG_PATH = _get_test_file_path("schedule_config_files/empty.yaml")
MINIMAL_CONFIG_PATH = _get_test_file_path("schedule_config_files/minimal.yaml")
FULL_CONFIG_PATH = _get_test_file_path("schedule_config_files/full.yaml")
UNRECOGNIZED_OPTION_CONFIG_PATH = _get_test_file_path(
    "schedule_config_files/unrecognized_option.yaml"
)

FULL_CONFIG_SCHEDULE_NAME = "test-name-from-file"


class FakeScheduleSDK:
    DEFAULT_SCHEDULE_NAME = "default-fake-schedule-name"

    def __init__(self):
        self.applied_config: Optional[ScheduleConfig] = None
        self.applied_id: Optional[str] = None
        self.applied_name: Optional[str] = None
        self._schedule_is_paused: Dict[str, bool] = {}
        self._schedule_trigger_counts: Dict[str, int] = defaultdict(int)

    def apply(self, config: ScheduleConfig):
        assert isinstance(config, ScheduleConfig)
        self.applied_config = config
        self.applied_id = str(uuid.uuid4())

        job_config = self.applied_config.job_config
        assert isinstance(job_config, JobConfig)
        self.applied_name = (
            job_config.name
            if job_config.name is not None
            else self.DEFAULT_SCHEDULE_NAME
        )
        self._schedule_is_paused[self.applied_id] = False
        self._schedule_is_paused[self.applied_name] = False
        return self.applied_id

    def set_state(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: ScheduleState,
    ) -> str:
        is_paused = state == ScheduleState.DISABLED
        if name is not None:
            self._schedule_is_paused[name] = is_paused
        elif id is not None:
            self._schedule_is_paused[id] = is_paused

        return id if id is not None else ""

    def schedule_is_paused(self, identifier: str) -> bool:
        return self._schedule_is_paused[identifier]

    def status(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ScheduleStatus:
        if (
            self.applied_config is not None
            and self.applied_name is not None
            and (name == self.applied_name or id == self.applied_id)
        ):
            if name is not None:
                is_paused = self._schedule_is_paused[name]
            elif id is not None:
                is_paused = self._schedule_is_paused[id]

            state = ScheduleState.DISABLED if is_paused else ScheduleState.ENABLED
            return ScheduleStatus(
                id=self.applied_id if self.applied_id else "",
                name=self.applied_name,
                state=state,
                config=self.applied_config,
            )
        raise RuntimeError("Schedule was not found.")

    def trigger(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        if name is not None:
            self._schedule_trigger_counts[name] += 1
        elif id is not None:
            self._schedule_trigger_counts[id] += 1

        return id if id is not None else ""

    def trigger_counts(self, identifier: str) -> int:
        return self._schedule_trigger_counts[identifier]


@pytest.fixture()
def fake_schedule_sdk() -> Generator[FakeScheduleSDK, None, None]:
    fake_schedule_sdk = FakeScheduleSDK()
    _LAZY_SDK_SINGLETONS[_SCHEDULE_SDK_SINGLETON_KEY] = fake_schedule_sdk
    try:
        yield fake_schedule_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_SCHEDULE_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestApply:
    def test_missing_arg(self, fake_schedule_sdk):
        runner = CliRunner()
        result = runner.invoke(apply)
        _assert_error_message(
            result, message="Error: Missing option '--config-file' / '-f'."
        )

    def test_config_file_not_found(self, fake_schedule_sdk):
        runner = CliRunner()
        result = runner.invoke(apply, ["-f", "missing_config.yaml"])
        _assert_error_message(
            result, message="Schedule config file 'missing_config.yaml' not found.",
        )

    @pytest.mark.parametrize(
        "config_file_arg", [MINIMAL_CONFIG_PATH, FULL_CONFIG_PATH],
    )
    def test_basic(self, fake_schedule_sdk, config_file_arg):
        runner = CliRunner()
        result = runner.invoke(apply, ["-f", config_file_arg])
        assert result.exit_code == 0, result.stdout
        assert fake_schedule_sdk.applied_config is not None

    def test_override_name(self, fake_schedule_sdk):
        runner = CliRunner()
        name = "test-different-name"
        result = runner.invoke(apply, ["--name", name, "-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert result.exit_code == 0, result.stdout
        assert fake_schedule_sdk.applied_config is not None
        assert fake_schedule_sdk.applied_name == name


class TestPauseResume:
    @pytest.mark.parametrize("command", [pause, resume])
    def test_no_identifiers(self, command, fake_schedule_sdk):
        runner = CliRunner()
        result = runner.invoke(command)
        _assert_error_message(
            result,
            message="One of '--name', '--id', or '--config-file' must be provided.",
        )

    @pytest.mark.parametrize("command", [pause, resume])
    @pytest.mark.parametrize(
        "identifiers",
        [
            ("name", "config_file", None),
            ("name", None, "id"),
            (None, "config_file", "id"),
            ("name", "config_file", "id"),
        ],
    )
    def test_multi_identifiers(self, command, identifiers, fake_schedule_sdk):
        runner = CliRunner()
        args = []
        name, config_file, id = identifiers  # noqa: A001
        if name:
            args += ["--name", name]
        if config_file:
            args += ["--config-file", config_file]
        if id:
            args += ["--id", id]

        result = runner.invoke(command, args)
        _assert_error_message(
            result,
            message="Only one of '--name', '--id', and '--config-file' can be provided.",
        )

    def test_name(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        assert not fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)

        result = runner.invoke(pause, ["--name", FULL_CONFIG_SCHEDULE_NAME])
        assert result.exit_code == 0
        assert fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)

        result = runner.invoke(resume, ["--name", FULL_CONFIG_SCHEDULE_NAME])
        assert result.exit_code == 0
        assert not fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)

    def test_id(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        assert not fake_schedule_sdk.schedule_is_paused(fake_schedule_sdk.applied_id)

        result = runner.invoke(pause, ["--id", fake_schedule_sdk.applied_id])
        assert result.exit_code == 0
        assert fake_schedule_sdk.schedule_is_paused(fake_schedule_sdk.applied_id)

        result = runner.invoke(resume, ["--id", fake_schedule_sdk.applied_id])
        assert result.exit_code == 0
        assert not fake_schedule_sdk.schedule_is_paused(fake_schedule_sdk.applied_id)

    def test_config_file(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        assert not fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)

        result = runner.invoke(pause, ["--config-file", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)

        result = runner.invoke(resume, ["--config-file", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert not fake_schedule_sdk.schedule_is_paused(FULL_CONFIG_SCHEDULE_NAME)


class TestStatus:
    def test_no_identifiers(self, fake_schedule_sdk):
        runner = CliRunner()
        result = runner.invoke(status)
        _assert_error_message(
            result,
            message="One of '--name', '--id', or '--config-file' must be provided.",
        )

    @pytest.mark.parametrize(
        "identifiers",
        [
            ("name", "config_file", None),
            ("name", None, "id"),
            (None, "config_file", "id"),
            ("name", "config_file", "id"),
        ],
    )
    def test_multi_identifiers(self, identifiers, fake_schedule_sdk):
        runner = CliRunner()
        args = []
        name, config_file, id = identifiers  # noqa: A001
        if name:
            args += ["--name", name]
        if config_file:
            args += ["--config-file", config_file]
        if id:
            args += ["--id", id]

        result = runner.invoke(status, args)
        _assert_error_message(
            result,
            message="Only one of '--name', '--id', and '--config-file' can be provided.",
        )

    def test_name(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(status, ["--name", FULL_CONFIG_SCHEDULE_NAME])
        assert result.exit_code == 0
        assert fake_schedule_sdk.applied_id in result.stdout
        assert fake_schedule_sdk.applied_name in result.stdout

    def test_id(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(status, ["--id", fake_schedule_sdk.applied_id])
        assert result.exit_code == 0
        assert fake_schedule_sdk.applied_id in result.stdout
        assert fake_schedule_sdk.applied_name in result.stdout

    def test_config_file(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(status, ["--config-file", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_schedule_sdk.applied_id in result.stdout
        assert fake_schedule_sdk.applied_name in result.stdout

    def test_verbose_flag(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])

        # No verbose flag -- exclude details
        result = runner.invoke(status, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert "id" in result.stdout
        assert "name" in result.stdout
        assert "state" in result.stdout
        assert "config" not in result.stdout

        # Verbose flag -- include details
        result = runner.invoke(status, ["-f", FULL_CONFIG_PATH, "-v"])
        assert result.exit_code == 0
        assert "id" in result.stdout
        assert "name" in result.stdout
        assert "state" in result.stdout
        assert "config" in result.stdout


class TestTrigger:
    def test_no_identifiers(self, fake_schedule_sdk):
        runner = CliRunner()
        result = runner.invoke(trigger)
        _assert_error_message(
            result,
            message="One of '--name', '--id', or '--config-file' must be provided.",
        )

    @pytest.mark.parametrize(
        "identifiers",
        [
            ("name", "config_file", None),
            ("name", None, "id"),
            (None, "config_file", "id"),
            ("name", "config_file", "id"),
        ],
    )
    def test_multi_identifiers(self, identifiers, fake_schedule_sdk):
        runner = CliRunner()
        args = []
        name, config_file, id = identifiers  # noqa: A001
        if name:
            args += ["--name", name]
        if config_file:
            args += ["--config-file", config_file]
        if id:
            args += ["--id", id]

        result = runner.invoke(trigger, args)
        _assert_error_message(
            result,
            message="Only one of '--name', '--id', and '--config-file' can be provided.",
        )

    def test_name(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(trigger, ["--name", FULL_CONFIG_SCHEDULE_NAME])
        assert result.exit_code == 0
        assert fake_schedule_sdk.trigger_counts(FULL_CONFIG_SCHEDULE_NAME) == 1

    def test_id(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(trigger, ["--id", fake_schedule_sdk.applied_id])
        assert result.exit_code == 0
        assert fake_schedule_sdk.trigger_counts(fake_schedule_sdk.applied_id) == 1

    def test_config_file(self, fake_schedule_sdk):
        runner = CliRunner()
        runner.invoke(apply, ["-f", FULL_CONFIG_PATH])
        result = runner.invoke(trigger, ["--config-file", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_schedule_sdk.trigger_counts(FULL_CONFIG_SCHEDULE_NAME) == 1



from io import StringIO
from json import dumps as json_dumps
import pathlib
from typing import Any, Dict, Optional, Tuple

import click
from typing_extensions import Literal
import yaml

from anyscale._private.models.image_uri import ImageURI
from anyscale.cli_logger import BlockLogger
from anyscale.commands.util import convert_kv_strings_to_dict
from anyscale.controllers.service_controller import ServiceController
import anyscale.service
from anyscale.service import ServiceConfig, ServiceState, ServiceStatus
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("service")
def service_cli():
    pass


def _read_name_from_config_file(path: str):
    """Read the 'name' property from the config file at `path`.

    This enables reading the name for both the existing (rollout) and new (deploy)
    config file formats.
    """
    if not pathlib.Path(path).is_file():
        raise click.ClickException(f"Config file not found at path: '{path}'.")

    with open(path) as f:
        config = yaml.safe_load(f)

    if config is None or "name" not in config:
        raise click.ClickException(f"No 'name' property found in config file '{path}'.")

    return config["name"]


@service_cli.command(
    name="deploy", help="Deploy or update a service.",
)
@click.argument("import_path", type=str, required=False, default=None)
@click.argument("arguments", nargs=-1, required=False)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to deploy. When deploying from a file, import path and arguments cannot be provided. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "-n",
    "--name",
    required=False,
    default=None,
    type=str,
    help="Unique name for the service. When running in a workspace, this defaults to the workspace name.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the service. This cannot be changed when using --in-place and is exclusive with --containerfile. When running in a workspace, this defaults to the image of the workspace.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the service. This cannot be changed when using --in-place and is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If not provided, the latest Ray version will be used.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the service. This cannot be changed when using --in-place. When running in a workspace, this defaults to the compute configuration of the workspace.",
)
@click.option(
    "-w",
    "--working-dir",
    required=False,
    default=None,
    type=str,
    help="Path to a local directory that will be the working directory for the service. The files in the directory will be automatically uploaded to cloud storage. When running in a workspace, this defaults to the current working directory.",
)
@click.option(
    "-e",
    "--exclude",
    required=False,
    type=str,
    multiple=True,
    help="File pattern to exclude when uploading local directories. This argument can be specified multiple times and the patterns will be appended to the 'excludes' list in the config file (if any).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the service. These will be installed on top of the image. When running in a workspace, this defaults to the workspace dependencies.",
)
@click.option(
    "-i",
    "--in-place",
    is_flag=True,
    show_default=True,
    default=False,
    help="Perform an in-place upgrade without starting a new cluster. This can be used for faster iteration during development but is *not* currently recommended for production deploys. This *cannot* be used to change cluster-level options such as image and compute config (they will be ignored).",
)
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic to send to the canary version of the service (0-100). This can be used to manually shift traffic toward (or away from) the canary version. If not provided, traffic will be shifted incrementally toward the canary version until it reaches 100. Not supported when using --in-place.",
)
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Amount of excess capacity allowed to be used while updating the service (0-100). Defaults to 100. Not supported when using --in-place.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the service. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
@click.option(
    "--py-module",
    required=False,
    default=None,
    multiple=True,
    type=str,
    help="Python modules to be available for import in the Ray workers. Each entry must be a path to a local directory.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def deploy(  # noqa: PLR0912, PLR0913 C901
    config_file: Optional[str],
    import_path: Optional[str],
    arguments: Tuple[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    working_dir: Optional[str],
    exclude: Tuple[str],
    requirements: Optional[str],
    in_place: bool,
    canary_percent: Optional[int],
    max_surge_percent: Optional[int],
    env: Optional[Tuple[str]],
    py_module: Tuple[str],
    cloud: Optional[str],
    project: Optional[str],
):
    """Deploy or update a service.

    If no service with the provided name is running, one will be created, else the existing service will be updated.

    To deploy a single application, the import path and optional arguments can be passed directly as arguments:

    `$ anyscale service deploy main:app arg1=val1`

    To deploy multiple applications, deploy from a config file using `-f`:

    `$ anyscale service deploy -f config.yaml`

    Command-line flags override values in the config file.
    """

    if config_file is not None:
        if import_path is not None or len(arguments) > 0:
            raise click.ClickException(
                "When a config file is provided, import path and application arguments can't be."
            )

        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = ServiceConfig.from_yaml(config_file)
    else:
        if import_path is None:
            raise click.ClickException(
                "Either config file or import path must be provided."
            )

        if (
            import_path.endswith((".yaml", ".yml"))
            or pathlib.Path(import_path).is_file()
        ):
            log.warning(
                f"The provided import path '{import_path}' looks like a config file. Did you mean to use '-f config.yaml'?"
            )

        app: Dict[str, Any] = {"import_path": import_path}
        arguments_dict = convert_kv_strings_to_dict(arguments)
        if arguments_dict:
            app["args"] = arguments_dict

        config = ServiceConfig(applications=[app])

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if working_dir is not None:
        config = config.options(working_dir=working_dir)

    if exclude:
        config = config.options(excludes=[e for e in exclude])

    if requirements is not None:
        config = config.options(requirements=requirements)

    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    if py_module:
        for module in py_module:
            if not pathlib.Path(module).is_dir():
                raise click.ClickException(
                    f"Python module path '{module}' does not exist or is not a directory."
                )
        config = config.options(py_modules=[*py_module])

    if cloud is not None:
        config = config.options(cloud=cloud)
    if project is not None:
        config = config.options(project=project)

    anyscale.service.deploy(
        config,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
    )


@service_cli.command(
    name="status", help="Get the status of a service.", hidden=True,
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to read the name from.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-j",
    "--json",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(
    name: Optional[str],
    config_file: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    json: bool,
    verbose: bool,
):
    if name is not None and config_file is not None:
        raise click.ClickException(
            "Only one of '--name' and '--config-file' can be provided."
        )

    if config_file is not None:
        name = _read_name_from_config_file(config_file)

    if name is None:
        raise click.ClickException(
            "Service name must be provided using '--name' or in a config file using '-f'."
        )

    status: ServiceStatus = anyscale.service.status(
        name=name, cloud=cloud, project=project
    )
    status_dict = status.to_dict()
    if not verbose:
        # TODO(edoakes): consider adding this as an API on the model itself if it
        # becomes a common pattern.
        status_dict.get("primary_version", {}).pop("config", None)
        status_dict.get("canary_version", {}).pop("config", None)

    if json:
        print(json_dumps(status_dict, indent=4, sort_keys=False))
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        print(stream.getvalue(), end="")


@service_cli.command(
    name="wait", help="Wait for a service to enter a target state.", hidden=True,
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to read the name from.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-s",
    "--state",
    default=str(ServiceState.RUNNING),
    help="The ServiceState to wait for the service to reach. Defaults to RUNNING.",
)
@click.option(
    "-t",
    "--timeout-s",
    default=600,
    type=float,
    help="Timeout to wait for the service to reach the target state. Defaults to 600s (10min).",
)
def wait(
    name: Optional[str],
    config_file: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    state: str,
    timeout_s: int,
):
    try:
        state = ServiceState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))

    if name is not None and config_file is not None:
        raise click.ClickException(
            "Only one of '--name' and '--config-file' can be provided."
        )

    if config_file is not None:
        name = _read_name_from_config_file(config_file)

    if name is None:
        raise click.ClickException(
            "Service name must be provided using '--name' or in a config file using '-f'."
        )

    anyscale.service.wait(
        name=name, cloud=cloud, project=project, state=state, timeout_s=timeout_s
    )


@service_cli.command(
    name="rollout", help="Roll out a service.",
)
@click.option(
    "-f",
    "--config-file",
    "--service-config-file",
    required=True,
    help="The path of the service configuration file.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option("--version", required=False, default=None, help="Version of service.")
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Max amount of excess capacity allocated during the rollout (0-100).",
)
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic to send to the canary version of the service.",
)
@click.option(
    "--rollout-strategy",
    required=False,
    default=None,
    type=click.Choice(["ROLLOUT", "IN_PLACE"], case_sensitive=False),
    help="Strategy for rollout.",
)
@click.option(
    "-i",
    "--in-place",
    "in_place",
    is_flag=True,
    show_default=True,
    default=False,
    help="Alias for `--rollout-strategy=IN_PLACE`.",
)
@click.option(
    "--no-auto-complete-rollout",
    is_flag=True,
    show_default=True,
    default=False,
    help="Do not complete the rollout (terminate the existing version cluster) after the canary version reaches 100%",
)
def rollout(  # noqa: PLR0913
    config_file: str,
    name: Optional[str],
    version: Optional[str],
    max_surge_percent: Optional[int],
    canary_percent: Optional[int],
    rollout_strategy: Optional[Literal["ROLLOUT", "IN_PLACE"]],
    in_place: bool,
    no_auto_complete_rollout: bool,
):
    """Start or update a service rollout to a new version."""
    if in_place:
        if rollout_strategy is not None:
            raise click.ClickException(
                "Only one of `--in-place/-i` and `--rollout-strategy` can be provided."
            )
        rollout_strategy = "IN_PLACE"

    service_controller = ServiceController()
    service_config = service_controller.read_service_config_file(config_file)
    service_controller.rollout(
        service_config,
        name=name,
        version=version,
        max_surge_percent=max_surge_percent,
        canary_percent=canary_percent,
        rollout_strategy=rollout_strategy,
        auto_complete_rollout=not no_auto_complete_rollout,
    )


# TODO(mowen): Add cloud support for this when we refactor to new SDK method
@service_cli.command(name="list", help="Display information about services.")
@click.option(
    "--name", "-n", required=False, default=None, help="Filter by service name."
)
@click.option(
    "--service-id", "--id", required=False, default=None, help="Filter by service id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--created-by-me",
    help="List services created by me only.",
    is_flag=True,
    default=False,
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001
    name: Optional[str],
    service_id: Optional[str],
    project_id: Optional[str],
    created_by_me: bool,
    max_items: int,
):
    """List services based on the provided filters.

    This returns both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_controller.list(
        name=name,
        service_id=service_id,
        project_id=project_id,
        created_by_me=created_by_me,
        max_items=max_items,
    )


# TODO(mowen): Add cloud support for this when we refactor to new SDK method
@service_cli.command(name="rollback", help="Roll back a service.")
@click.option(
    "--service-id", "--id", default=None, help="ID of service.",
)
@click.option("-n", "--name", required=False, default=None, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "-f",
    "--config-file",
    "--service-config-file",
    help="Path to a YAML config file to read the name from. `--service-config-file` is deprecated, use `-f` or `--config-file`.",
)
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Max amount of excess capacity allocated during the rollback (0-100).",
)
def rollback(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    config_file: Optional[str],
    max_surge_percent: Optional[int],
):
    """Perform a rollback for a service that is currently in a rollout."""
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=config_file,
        project_id=project_id,
    )
    service_controller.rollback(service_id, max_surge_percent)


# TODO(mowen): Add cloud support for this when we refactor to new SDK method
@service_cli.command(name="terminate", help="Terminate a service.")
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option("-n", "--name", required=False, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "-f",
    "--config-file",
    "--service-config-file",
    help="Path to a YAML config file to read the name from. `--service-config-file` is deprecated, use `-f` or `--config-file`.",
)
def terminate(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    config_file: Optional[str],
):
    """Terminate a service.

    This applies to both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=config_file,
        project_id=project_id,
    )
    service_controller.terminate(service_id)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class MachineInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'machine_id': 'str',
        'hostname': 'str',
        'machine_shape': 'str',
        'num_cpus': 'int',
        'num_gpus': 'int',
        'memory_bytes': 'int',
        'connection_state': 'MachineConnectionState',
        'allocation_state': 'MachineAllocationState',
        'node_id': 'str',
        'cluster_id': 'str',
        'cluster_creator_name': 'str',
        'cluster_name': 'str',
        'anyscale_version': 'str'
    }

    attribute_map = {
        'machine_id': 'machine_id',
        'hostname': 'hostname',
        'machine_shape': 'machine_shape',
        'num_cpus': 'num_cpus',
        'num_gpus': 'num_gpus',
        'memory_bytes': 'memory_bytes',
        'connection_state': 'connection_state',
        'allocation_state': 'allocation_state',
        'node_id': 'node_id',
        'cluster_id': 'cluster_id',
        'cluster_creator_name': 'cluster_creator_name',
        'cluster_name': 'cluster_name',
        'anyscale_version': 'anyscale_version'
    }

    def __init__(self, machine_id=None, hostname=None, machine_shape=None, num_cpus=None, num_gpus=None, memory_bytes=None, connection_state=None, allocation_state=None, node_id=None, cluster_id=None, cluster_creator_name=None, cluster_name=None, anyscale_version=None, local_vars_configuration=None):  # noqa: E501
        """MachineInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._machine_id = None
        self._hostname = None
        self._machine_shape = None
        self._num_cpus = None
        self._num_gpus = None
        self._memory_bytes = None
        self._connection_state = None
        self._allocation_state = None
        self._node_id = None
        self._cluster_id = None
        self._cluster_creator_name = None
        self._cluster_name = None
        self._anyscale_version = None
        self.discriminator = None

        self.machine_id = machine_id
        self.hostname = hostname
        self.machine_shape = machine_shape
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.memory_bytes = memory_bytes
        self.connection_state = connection_state
        self.allocation_state = allocation_state
        if node_id is not None:
            self.node_id = node_id
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if cluster_creator_name is not None:
            self.cluster_creator_name = cluster_creator_name
        if cluster_name is not None:
            self.cluster_name = cluster_name
        if anyscale_version is not None:
            self.anyscale_version = anyscale_version

    @property
    def machine_id(self):
        """Gets the machine_id of this MachineInfo.  # noqa: E501

        The ID for this machine.  # noqa: E501

        :return: The machine_id of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._machine_id

    @machine_id.setter
    def machine_id(self, machine_id):
        """Sets the machine_id of this MachineInfo.

        The ID for this machine.  # noqa: E501

        :param machine_id: The machine_id of this MachineInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and machine_id is None:  # noqa: E501
            raise ValueError("Invalid value for `machine_id`, must not be `None`")  # noqa: E501

        self._machine_id = machine_id

    @property
    def hostname(self):
        """Gets the hostname of this MachineInfo.  # noqa: E501

        The hostname for this machine.  # noqa: E501

        :return: The hostname of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._hostname

    @hostname.setter
    def hostname(self, hostname):
        """Sets the hostname of this MachineInfo.

        The hostname for this machine.  # noqa: E501

        :param hostname: The hostname of this MachineInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and hostname is None:  # noqa: E501
            raise ValueError("Invalid value for `hostname`, must not be `None`")  # noqa: E501

        self._hostname = hostname

    @property
    def machine_shape(self):
        """Gets the machine_shape of this MachineInfo.  # noqa: E501

        The physical shape for this machine.  # noqa: E501

        :return: The machine_shape of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._machine_shape

    @machine_shape.setter
    def machine_shape(self, machine_shape):
        """Sets the machine_shape of this MachineInfo.

        The physical shape for this machine.  # noqa: E501

        :param machine_shape: The machine_shape of this MachineInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and machine_shape is None:  # noqa: E501
            raise ValueError("Invalid value for `machine_shape`, must not be `None`")  # noqa: E501

        self._machine_shape = machine_shape

    @property
    def num_cpus(self):
        """Gets the num_cpus of this MachineInfo.  # noqa: E501

        The number of CPUs for this machine.  # noqa: E501

        :return: The num_cpus of this MachineInfo.  # noqa: E501
        :rtype: int
        """
        return self._num_cpus

    @num_cpus.setter
    def num_cpus(self, num_cpus):
        """Sets the num_cpus of this MachineInfo.

        The number of CPUs for this machine.  # noqa: E501

        :param num_cpus: The num_cpus of this MachineInfo.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_cpus is None:  # noqa: E501
            raise ValueError("Invalid value for `num_cpus`, must not be `None`")  # noqa: E501

        self._num_cpus = num_cpus

    @property
    def num_gpus(self):
        """Gets the num_gpus of this MachineInfo.  # noqa: E501

        The number of GPUs for this machine.  # noqa: E501

        :return: The num_gpus of this MachineInfo.  # noqa: E501
        :rtype: int
        """
        return self._num_gpus

    @num_gpus.setter
    def num_gpus(self, num_gpus):
        """Sets the num_gpus of this MachineInfo.

        The number of GPUs for this machine.  # noqa: E501

        :param num_gpus: The num_gpus of this MachineInfo.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_gpus is None:  # noqa: E501
            raise ValueError("Invalid value for `num_gpus`, must not be `None`")  # noqa: E501

        self._num_gpus = num_gpus

    @property
    def memory_bytes(self):
        """Gets the memory_bytes of this MachineInfo.  # noqa: E501

        The amount of memory in bytes for this machine.  # noqa: E501

        :return: The memory_bytes of this MachineInfo.  # noqa: E501
        :rtype: int
        """
        return self._memory_bytes

    @memory_bytes.setter
    def memory_bytes(self, memory_bytes):
        """Sets the memory_bytes of this MachineInfo.

        The amount of memory in bytes for this machine.  # noqa: E501

        :param memory_bytes: The memory_bytes of this MachineInfo.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and memory_bytes is None:  # noqa: E501
            raise ValueError("Invalid value for `memory_bytes`, must not be `None`")  # noqa: E501

        self._memory_bytes = memory_bytes

    @property
    def connection_state(self):
        """Gets the connection_state of this MachineInfo.  # noqa: E501

        The connection state for this machine.  # noqa: E501

        :return: The connection_state of this MachineInfo.  # noqa: E501
        :rtype: MachineConnectionState
        """
        return self._connection_state

    @connection_state.setter
    def connection_state(self, connection_state):
        """Sets the connection_state of this MachineInfo.

        The connection state for this machine.  # noqa: E501

        :param connection_state: The connection_state of this MachineInfo.  # noqa: E501
        :type: MachineConnectionState
        """
        if self.local_vars_configuration.client_side_validation and connection_state is None:  # noqa: E501
            raise ValueError("Invalid value for `connection_state`, must not be `None`")  # noqa: E501

        self._connection_state = connection_state

    @property
    def allocation_state(self):
        """Gets the allocation_state of this MachineInfo.  # noqa: E501

        The allocation state for this machine.  # noqa: E501

        :return: The allocation_state of this MachineInfo.  # noqa: E501
        :rtype: MachineAllocationState
        """
        return self._allocation_state

    @allocation_state.setter
    def allocation_state(self, allocation_state):
        """Sets the allocation_state of this MachineInfo.

        The allocation state for this machine.  # noqa: E501

        :param allocation_state: The allocation_state of this MachineInfo.  # noqa: E501
        :type: MachineAllocationState
        """
        if self.local_vars_configuration.client_side_validation and allocation_state is None:  # noqa: E501
            raise ValueError("Invalid value for `allocation_state`, must not be `None`")  # noqa: E501

        self._allocation_state = allocation_state

    @property
    def node_id(self):
        """Gets the node_id of this MachineInfo.  # noqa: E501

        The node ID for this machine if it's allocated to a Ray cluster.  # noqa: E501

        :return: The node_id of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._node_id

    @node_id.setter
    def node_id(self, node_id):
        """Sets the node_id of this MachineInfo.

        The node ID for this machine if it's allocated to a Ray cluster.  # noqa: E501

        :param node_id: The node_id of this MachineInfo.  # noqa: E501
        :type: str
        """

        self._node_id = node_id

    @property
    def cluster_id(self):
        """Gets the cluster_id of this MachineInfo.  # noqa: E501

        If allocated to a Ray cluster, the ID of the cluster.  # noqa: E501

        :return: The cluster_id of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this MachineInfo.

        If allocated to a Ray cluster, the ID of the cluster.  # noqa: E501

        :param cluster_id: The cluster_id of this MachineInfo.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def cluster_creator_name(self):
        """Gets the cluster_creator_name of this MachineInfo.  # noqa: E501

        If allocated to a Ray cluster, the name of the creator of the cluster.  # noqa: E501

        :return: The cluster_creator_name of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._cluster_creator_name

    @cluster_creator_name.setter
    def cluster_creator_name(self, cluster_creator_name):
        """Sets the cluster_creator_name of this MachineInfo.

        If allocated to a Ray cluster, the name of the creator of the cluster.  # noqa: E501

        :param cluster_creator_name: The cluster_creator_name of this MachineInfo.  # noqa: E501
        :type: str
        """

        self._cluster_creator_name = cluster_creator_name

    @property
    def cluster_name(self):
        """Gets the cluster_name of this MachineInfo.  # noqa: E501

        If allocated to a Ray cluster, the name of the cluster.  # noqa: E501

        :return: The cluster_name of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._cluster_name

    @cluster_name.setter
    def cluster_name(self, cluster_name):
        """Sets the cluster_name of this MachineInfo.

        If allocated to a Ray cluster, the name of the cluster.  # noqa: E501

        :param cluster_name: The cluster_name of this MachineInfo.  # noqa: E501
        :type: str
        """

        self._cluster_name = cluster_name

    @property
    def anyscale_version(self):
        """Gets the anyscale_version of this MachineInfo.  # noqa: E501

        The version of Anyscale running on this machine.  # noqa: E501

        :return: The anyscale_version of this MachineInfo.  # noqa: E501
        :rtype: str
        """
        return self._anyscale_version

    @anyscale_version.setter
    def anyscale_version(self, anyscale_version):
        """Sets the anyscale_version of this MachineInfo.

        The version of Anyscale running on this machine.  # noqa: E501

        :param anyscale_version: The anyscale_version of this MachineInfo.  # noqa: E501
        :type: str
        """

        self._anyscale_version = anyscale_version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MachineInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MachineInfo):
            return True

        return self.to_dict() != other.to_dict()



from typing import Optional

import click


@click.group("dataset", help="Dataset files stored on your Anyscale cloud.")
def dataset_cli():
    pass


@dataset_cli.command(
    name="upload", short_help="Upload a dataset to your Anyscale cloud.",
)
@click.argument("dataset_file", required=True)
@click.option(
    "--existing-dataset-id-or-name",
    "-d",
    required=False,
    help="ID or name of an existing dataset to upload a new version of. "
    "If not provided, a new dataset will be created.",
)
@click.option(
    "--description",
    required=False,
    help="Description of the dataset (or dataset version).",
)
@click.option(
    "--cloud-id",
    required=False,
    help="ID of the Anyscale cloud to upload a new dataset to. "
    "If not provided, the default Anyscale Hosted cloud will be used.",
)
@click.option(
    "--project-id",
    required=False,
    help="ID of the Anyscale project to upload a new dataset to. "
    "If not provided, the default project of the cloud will be used.",
)
def upload_dataset(
    dataset_file: str,
    existing_dataset_id_or_name: Optional[str],
    description: Optional[str],
    cloud_id: Optional[str],
    project_id: Optional[str],
):
    """
    Uploads a dataset, or a new version of a dataset, to your Anyscale cloud.

    DATASET_FILE = Path to the dataset file to upload

    Example usage:

        anyscale llm dataset upload path/to/my_dataset.jsonl

        anyscale llm dataset upload my_dataset2.jsonl -d my_dataset.jsonl

        anyscale llm dataset upload my_dataset2.jsonl -d file_123 --description 'added 3 lines'

    \b
    NOTE:
    If you are uploading a new version, have run this from within an Anyscale workspace,
    and neither `--cloud-id` nor `--project-id` is provided, the cloud and project of the workspace will be used.
    """
    print(f"Uploading dataset: {dataset_file}")
    print(f"Existing dataset ID: {existing_dataset_id_or_name}")
    print(f"Description: {description}")
    print(f"Cloud ID: {cloud_id}")
    print(f"Project ID: {project_id}")


@dataset_cli.command(
    name="download", short_help="Download a dataset.",
)
@click.argument(
    "dataset_id_or_name", required=True,
)
@click.option(
    "--version",
    "-v",
    required=False,
    type=int,
    help="Version of the dataset to download. "
    "If a negative integer is provided, the dataset returned is this many versions back of the latest version. "
    "Default: Latest version.",
)
def download_dataset(
    dataset_id_or_name: str, version: Optional[int],
):
    """
    Downloads a dataset from your Anyscale cloud.

    DATASET_ID_OR_NAME = ID or name of the dataset to download

    Example usage:

        anyscale llm dataset download my_dataset.jsonl

        anyscale llm dataset download file_123

    Retrieve the second latest version of the dataset:

        anyscale llm dataset download my_dataset.jsonl -v -1
    """
    print(f"Downloading dataset: {dataset_id_or_name}")
    print(f"Version: {version}")



from typing import Optional


def upload(
    dataset_file: str,  # noqa: ARG001
    existing_dataset_id_or_name: Optional[str] = None,  # noqa: ARG001
    *,
    description: Optional[str] = None,  # noqa: ARG001
    cloud_id: Optional[str] = None,  # noqa: ARG001
    project_id: Optional[str] = None,  # noqa: ARG001
):
    """Uploads a dataset, or a new version of a dataset, to your Anyscale cloud.

    :param dataset_file: Path to the dataset file to upload.
    :param existing_dataset_id_or_name: ID or name of an existing dataset to upload a new version of. If not provided, a new dataset will be created.
    :param description: Description of the dataset (or dataset version).
    :param cloud_id: ID of the Anyscale cloud to upload a new dataset to. If not provided, the default Anyscale Hosted cloud will be used.
    :param project_id: ID of the Anyscale project to upload a new dataset to. If not provided, the default project of the cloud will be used.

    Example usage:
    ```python
    anyscale.llm.dataset.upload("path/to/my_dataset.jsonl")
    anyscale.llm.dataset.upload("my_dataset2.jsonl", "file_123")
    anyscale.llm.dataset.upload("my_dataset2.jsonl", "my_dataset.jsonl", description="added 3 lines")
    ```
    :return: The `Dataset` object representing the uploaded dataset.

    NOTE:
    If you are uploading a new version, have run this from within an Anyscale workspace,
    and neither `cloud_id` nor `project_id` are provided, the cloud and project of the workspace will be used.
    """
    raise NotImplementedError


def download(
    dataset_id_or_name: str, version: Optional[int] = None  # noqa: ARG001
) -> bytes:
    """Downloads a dataset from your Anyscale cloud.

    :param dataset_id_or_name: ID or name of the dataset to download.
    :param version: Version of the dataset to download. If a negative integer is provided, the dataset returned is this many versions back of the latest version. Default: Latest version.

    Example usage:
    ```python
    dataset_contents = anyscale.llm.dataset.download("my_dataset.jsonl") # anyscale.llm.dataset.download("file_123")
    jsonl_obj = [json.loads(line) for line in dataset_contents.decode().splitlines()]

    prev_dataset_contents = anyscale.llm.dataset.download("my_dataset.jsonl", version=-1)
    ```
    :return: The contents of the dataset file.
    """
    raise NotImplementedError



from abc import ABC, abstractmethod
import io
import logging
import os
import pathlib
import time
from typing import Any, Callable, Dict, Generator, IO, List, Optional, Tuple
from urllib.parse import urlparse

from openapi_client.exceptions import ApiException
import requests

from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface,
    DEFAULT_PYTHON_VERSION,
    DEFAULT_RAY_VERSION,
    RUNTIME_ENV_PACKAGE_FORMAT,
)
from anyscale._private.models.image_uri import ImageURI
from anyscale.api_utils.logs_util import _download_log_from_s3_url_sync
from anyscale.authenticate import AuthenticationBlock, get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models import (
    ArchiveStatus,
    Cloud,
    CloudDataBucketFileType,
    CloudDataBucketPresignedUploadInfo,
    CloudDataBucketPresignedUploadRequest,
    CloudNameOptions,
    ComputeTemplate,
    ComputeTemplateConfig,
    ComputeTemplateQuery,
    CreateComputeTemplate,
    CreateInternalProductionJob,
    DecoratedComputeTemplate,
    InternalProductionJob,
)
from anyscale.client.openapi_client.models.create_schedule import CreateSchedule
from anyscale.client.openapi_client.models.decorated_schedule import DecoratedSchedule
from anyscale.client.openapi_client.models.production_job import ProductionJob
from anyscale.client.openapi_client.rest import ApiException as InternalApiException
from anyscale.cluster_compute import parse_cluster_compute_name_version
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as ExternalApi
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cluster,
    ClusterCompute,
    ClusterComputeConfig,
    ClusterEnvironment,
    ClusterEnvironmentBuild,
    ClusterenvironmentbuildListResponse,
    ClusterEnvironmentBuildStatus,
    ClusterEnvironmentsQuery,
    CreateClusterEnvironment,
    CreateClusterEnvironmentBuild,
    Job as APIJobRun,
    Project,
    RollbackServiceModel,
    ServiceModel,
    TextQuery,
)
from anyscale.sdk.anyscale_client.models.jobs_query import JobsQuery
from anyscale.sdk.anyscale_client.models.jobs_sort_field import JobsSortField
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.sort_by_clause_jobs_sort_field import (
    SortByClauseJobsSortField,
)
from anyscale.sdk.anyscale_client.models.sort_order import SortOrder
from anyscale.sdk.anyscale_client.rest import ApiException as ExternalApiException
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST
from anyscale.shared_anyscale_utils.latest_ray_version import LATEST_RAY_VERSION
from anyscale.util import (
    get_cluster_model_for_current_workspace,
    get_endpoint,
    is_anyscale_workspace,
)
from anyscale.utils.connect_helpers import search_entities
from anyscale.utils.runtime_env import (
    is_workspace_dependency_tracking_disabled,
    parse_dot_env_file,
    WORKSPACE_REQUIREMENTS_FILE_PATH,
    zip_local_dir,
)
from anyscale.utils.workspace_notification import (
    WORKSPACE_NOTIFICATION_ADDRESS,
    WorkspaceNotification,
)


WORKSPACE_ID_ENV_VAR = "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"
OVERWRITE_EXISTING_CLOUD_STORAGE_FILES = (
    os.environ.get("ANYSCALE_OVERWRITE_EXISTING_CLOUD_STORAGE_FILES", "0") == "1"
)

# internal_logger is used for logging internal errors or debug messages that we do not expect users to see.
internal_logger = logging.getLogger(__name__)


class AWSS3ClientInterface(ABC):
    @abstractmethod
    def download_fileobj(self, Bucket: str, Key: str, Fileobj: IO[Any],) -> None:
        """Download a file from an S3 bucket to a file-like object."""
        raise NotImplementedError


class GCSBlobInterface(ABC):
    @abstractmethod
    def download_to_file(self, fileobj: IO[Any]) -> None:
        """Download the blob to a file-like object."""
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        """Check if the blob exists."""
        raise NotImplementedError


class GCSBucketInterface(ABC):
    @abstractmethod
    def blob(self, object_name: str) -> GCSBlobInterface:
        """Get a blob object for the given object name."""
        raise NotImplementedError


class GCPGCSClientInterface(ABC):
    @abstractmethod
    def bucket(self, bucket: str) -> GCSBucketInterface:
        """Get a bucket object for the given bucket name."""
        raise NotImplementedError


class AnyscaleClient(AnyscaleClientInterface):
    # Number of entries to fetch per request for list endpoints.
    LIST_ENDPOINT_COUNT = 50

    def __init__(
        self,
        *,
        api_clients: Optional[Tuple[ExternalApi, InternalApi]] = None,
        sleep: Optional[Callable[[float], None]] = None,
        workspace_requirements_file_path: str = WORKSPACE_REQUIREMENTS_FILE_PATH,
        logger: Optional[BlockLogger] = None,
        host: Optional[str] = None,
        s3_client: Optional[AWSS3ClientInterface] = None,
        gcs_client: Optional[GCPGCSClientInterface] = None,
    ):
        if api_clients is None:
            auth_block: AuthenticationBlock = get_auth_api_client(
                raise_structured_exception=True
            )
            api_clients = (auth_block.anyscale_api_client, auth_block.api_client)

        self._external_api_client, self._internal_api_client = api_clients
        self._workspace_requirements_file_path = workspace_requirements_file_path
        self._sleep = sleep or time.sleep
        self._s3_client = s3_client
        self._gcs_client = gcs_client

        # Cached IDs and models to avoid duplicate lookups.
        self._default_project_id_from_cloud_id: Dict[Optional[str], str] = {}
        self._cloud_id_cache: Dict[Optional[str], str] = {}
        self._current_workspace_cluster: Optional[Cluster] = None
        self._logger = logger or BlockLogger()
        self._host = host or ANYSCALE_HOST

    @property
    def s3_client(self) -> AWSS3ClientInterface:
        if self._s3_client is None:
            # initialize the s3 client lazily so that we import the boto3 library only when needed.
            try:
                import boto3
                import botocore.config
            except ImportError:
                raise RuntimeError(
                    "Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
                )
            self._s3_client = boto3.client(  # type: ignore
                "s3", config=botocore.config.Config(signature_version="s3v4")
            )
        return self._s3_client  # type: ignore

    @property
    def gcs_client(self) -> GCPGCSClientInterface:
        if self._gcs_client is None:
            # initialize the gcs client lazily so that we import the google cloud storage library only when needed.
            try:
                from google.cloud import storage
            except ImportError:
                raise RuntimeError(
                    "Could not import the Google Storage Python API via `from google.cloud import storage`.  Please check your installation or try running `pip install --upgrade google-cloud-storage`."
                )
            self._gcs_client = storage.Client()
        return self._gcs_client  # type: ignore

    @property
    def host(self) -> str:
        return self._host

    @property
    def logger(self) -> BlockLogger:
        return self._logger

    def get_job_ui_url(self, job_id: str) -> str:
        return get_endpoint(f"/jobs/{job_id}", host=self.host)

    def get_service_ui_url(self, service_id: str) -> str:
        return get_endpoint(f"/services/{service_id}", host=self.host)

    def get_compute_config_ui_url(
        self, compute_config_id: str, *, cloud_id: str
    ) -> str:
        return get_endpoint(
            f"/v2/{cloud_id}/compute-configs/{compute_config_id}", host=self.host
        )

    def get_build_ui_url(self, cluster_env_id, build_id: str) -> str:
        return get_endpoint(
            f"v2/container-images/{cluster_env_id}/versions/{build_id}", host=self.host
        )

    def get_current_workspace_id(self) -> Optional[str]:
        return os.environ.get(WORKSPACE_ID_ENV_VAR, None)

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_id() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        if (
            not self.inside_workspace()
            or is_workspace_dependency_tracking_disabled()
            or not pathlib.Path(self._workspace_requirements_file_path).is_file()
        ):
            return None

        return self._workspace_requirements_file_path

    def _download_file_from_google_cloud_storage(
        self, bucket: str, object_name: str
    ) -> Optional[bytes]:
        try:
            bucket_obj = self.gcs_client.bucket(bucket)
            blob = bucket_obj.blob(object_name)
            fileobj = io.BytesIO()
            if blob.exists():
                blob.download_to_file(fileobj)
                return fileobj.getvalue()
            return None
        except Exception as e:  # noqa: BLE001
            raise RuntimeError(
                f"Failed to download the working directory from Google Cloud Storage. Error {e!r}"
                "Please validate you have exported cloud credentials with the correct read permissions and the intended bucket exists in your Cloud Storage account. "
            ) from e

    def _download_file_from_s3(self, bucket: str, object_key: str) -> Optional[bytes]:
        try:
            from botocore.exceptions import ClientError
        except Exception:  # noqa: BLE001
            raise RuntimeError(
                "Could not download file from S3: Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
            )
        try:
            fileobj = io.BytesIO()
            self.s3_client.download_fileobj(bucket, object_key, fileobj)
            return fileobj.getvalue()
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return None
            raise
        except Exception as e:  # noqa: BLE001
            raise RuntimeError(
                f"Failed to download the working directory from S3. Error {e!r}"
                "Please validate you have exported cloud credentials with the correct read permissions and the intended bucket exists in your S3 account. "
            ) from e

    def _download_file_from_remote_storage(self, remote_uri: str) -> Optional[bytes]:
        parsed_uri = urlparse(remote_uri)
        service = parsed_uri.scheme
        bucket = parsed_uri.netloc
        object_name = parsed_uri.path.lstrip("/")
        if service == "s3":
            return self._download_file_from_s3(bucket, object_name)
        if service == "gs":
            return self._download_file_from_google_cloud_storage(bucket, object_name)
        return None

    def get_workspace_env_vars(self) -> Optional[Dict[str, str]]:
        system_storage_path = os.environ.get("ANYSCALE_INTERNAL_SYSTEM_STORAGE", "")
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID", "")
        workspace_artifact_path = (
            os.path.join(
                system_storage_path, "workspace_tracking_dependencies", workspace_id,
            )
            if workspace_id and system_storage_path
            else ""
        )
        workspace_dot_env_path = (
            os.path.join(workspace_artifact_path, ".env")
            if workspace_artifact_path
            else ""
        )

        if not self.inside_workspace() or not workspace_dot_env_path:
            return None

        dot_env = self._download_file_from_remote_storage(workspace_dot_env_path)
        if dot_env:
            parsed_dot_env = parse_dot_env_file(dot_env)
            if parsed_dot_env:
                self.logger.info(
                    f"Using workspace runtime dependencies env vars: {parsed_dot_env}."
                )
            return parsed_dot_env
        return None

    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        # Checks for the existence of the ANYSCALE_EXPERIMENTAL_WORKSPACE_ID env var.
        if not is_anyscale_workspace():
            return None

        if self._current_workspace_cluster is None:
            # Picks up the cluster ID from the ANYSCALE_SESSION_ID env var.
            self._current_workspace_cluster = get_cluster_model_for_current_workspace(
                self._external_api_client
            )

        return self._current_workspace_cluster

    def _get_project_id_by_name(
        self, *, parent_cloud_id: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        # Find if project with name already exists
        existing_projects = self._internal_api_client.find_project_by_project_name_api_v2_projects_find_by_name_get(
            name
        ).results
        if len(existing_projects) == 0:
            raise ValueError(f"Project '{name}' was not found.")
        else:
            for project in existing_projects:
                if project.parent_cloud_id == parent_cloud_id:
                    return project.id
            raise ValueError(
                f"{len(existing_projects)} project(s) found with name '{name}' and none matched cloud_id '{parent_cloud_id}'"
            )

    def _get_project_id_by_cloud_id(
        self, *, parent_cloud_id: Optional[str] = None,
    ) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            if (
                workspace_cluster.cluster_compute_config is not None
                and workspace_cluster.cluster_compute_config.cloud_id == parent_cloud_id
            ):
                return workspace_cluster.project_id
            elif workspace_cluster.cluster_compute_id is not None:
                workspace_cluster_compute = self.get_compute_config(
                    workspace_cluster.cluster_compute_id
                )
                if (
                    workspace_cluster_compute is not None
                    and workspace_cluster_compute.config is not None
                    and workspace_cluster_compute.config.cloud_id == parent_cloud_id
                ):
                    return workspace_cluster.project_id

        if self._default_project_id_from_cloud_id.get(parent_cloud_id) is None:
            # Cloud isolation organizations follow the permissions model in https://docs.anyscale.com/organization-and-user-account/access-controls
            # TODO(nikita): Remove this FF check after completing the cloud isolation migration in Q2 2024
            cloud_isolation_ff_on = self._internal_api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
                "cloud-isolation-phase-1"
            ).result.is_on
            default_project: Project = self._external_api_client.get_default_project(
                parent_cloud_id=(parent_cloud_id if cloud_isolation_ff_on else None)
            ).result
            self._default_project_id_from_cloud_id[parent_cloud_id] = default_project.id

        return self._default_project_id_from_cloud_id[parent_cloud_id]

    def get_project_id(
        self, *, parent_cloud_id: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        if name is not None:
            return self._get_project_id_by_name(
                parent_cloud_id=parent_cloud_id, name=name
            )
        else:
            return self._get_project_id_by_cloud_id(parent_cloud_id=parent_cloud_id)

    def _get_cloud_id_for_compute_config_id(self, compute_config_id: str) -> str:
        cluster_compute: ClusterCompute = self._external_api_client.get_cluster_compute(
            compute_config_id
        ).result
        cluster_compute_config: ClusterComputeConfig = cluster_compute.config
        return cluster_compute_config.cloud_id

    def _get_cloud_id_by_name(self, cloud_name: str) -> Optional[str]:
        try:
            return self._internal_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post(
                CloudNameOptions(name=cloud_name),
            ).result.id
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_cloud_id(
        self, cloud_name: Optional[str] = None, compute_config_id: Optional[str] = None
    ) -> str:
        if cloud_name is not None and compute_config_id is not None:
            raise ValueError(
                "Only one of cloud_name or compute_config_id should be provided."
            )

        if compute_config_id is not None:
            return self._get_cloud_id_for_compute_config_id(compute_config_id)

        if cloud_name in self._cloud_id_cache:
            return self._cloud_id_cache[cloud_name]

        if cloud_name is not None:
            cloud_id = self._get_cloud_id_by_name(cloud_name)
            if cloud_id is None:
                raise RuntimeError(f"Cloud '{cloud_name}' not found.")
        elif self.inside_workspace():
            workspace_cluster = self.get_current_workspace_cluster()
            assert workspace_cluster is not None
            # NOTE(edoakes): the Cluster model has a compute_config_config model that includes
            # its cloud ID, but it's not always populated.
            # TODO(edoakes): add cloud_id to the Cluster model to avoid a second RTT.
            cloud_id = self._get_cloud_id_for_compute_config_id(
                workspace_cluster.cluster_compute_id
            )
        else:
            cloud_id = self._external_api_client.get_default_cloud().result.id

        assert cloud_id is not None
        self._cloud_id_cache[cloud_name] = cloud_id
        return cloud_id

    def get_cloud(self, *, cloud_id: str) -> Optional[Cloud]:
        try:
            cloud: Cloud = self._internal_api_client.get_cloud_api_v2_clouds_cloud_id_get(
                cloud_id
            ).result
            return cloud
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def create_compute_config(
        self, config: ComputeTemplateConfig, *, name: Optional[str] = None
    ) -> Tuple[str, str]:
        result: ComputeTemplate = self._internal_api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                config=config, name=name, anonymous=name is None, new_version=True
            )
        ).result
        return f"{result.name}:{result.version}", result.id

    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        try:
            return self._internal_api_client.get_compute_template_api_v2_compute_templates_template_id_get(
                compute_config_id
            ).result
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_compute_config_id(
        self,
        compute_config_name: Optional[str] = None,
        *,
        include_archived: bool = False,
    ) -> Optional[str]:
        if compute_config_name is not None:
            name, version = parse_cluster_compute_name_version(compute_config_name)
            if version is None:
                # Setting `version=-1` will return only the latest version if there are multiple.
                version = -1
            cluster_computes = self._internal_api_client.search_compute_templates_api_v2_compute_templates_search_post(
                ComputeTemplateQuery(
                    orgwide=True,
                    name={"equals": name},
                    include_anonymous=True,
                    archive_status=ArchiveStatus.ALL
                    if include_archived
                    else ArchiveStatus.NOT_ARCHIVED,
                    version=version,
                )
            ).results

            if len(cluster_computes) == 0:
                return None

            compute_template: DecoratedComputeTemplate = cluster_computes[0]
            return compute_template.id

        # If the compute config name is not provided, we pick an appropriate default.
        #
        #   - If running in a workspace:
        #       * If auto_select_worker_config enabled: we switch over to a standardized
        #         default compute config (copying over any cluster-level attributes, e.g.
        #         max-gpus).
        #       * Otherwise, we use the workspace's compute config.
        #
        #   - Otherwise, we use the default compute config provided by the API.

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            workspace_compute_config: DecoratedComputeTemplate = self.get_compute_config(
                workspace_cluster.cluster_compute_id
            )
            workspace_config: ClusterComputeConfig = workspace_compute_config.config
            if workspace_config.auto_select_worker_config:
                standard_template = self._build_standard_compute_template_from_existing_auto_config(
                    workspace_config
                )
                _, compute_config_id = self.create_compute_config(standard_template)
                return compute_config_id
            else:
                return workspace_cluster.cluster_compute_id

        return self.get_default_compute_config(cloud_id=self.get_cloud_id()).id

    def archive_compute_config(self, *, compute_config_id):
        self._internal_api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post(
            compute_config_id
        )

    def get_default_compute_config(self, *, cloud_id: str) -> ClusterCompute:
        return self._external_api_client.get_default_cluster_compute(
            cloud_id=cloud_id,
        ).result

    def _build_standard_compute_template_from_existing_auto_config(
        self, compute_config: ClusterComputeConfig
    ) -> ComputeTemplateConfig:
        """
        Build a standard compute template config from an existing compute config.

        1. Pull the default compute template config.
        2. Disable scheduling on the head node.
        3. Enable auto_select_worker_config.
        4. Copy any cluster-level flags from the provided compute config to the template.
        """
        # Retrieve the default cluster compute config for the cloud.
        default_compute_template: DecoratedComputeTemplate = self._external_api_client.get_default_cluster_compute(
            cloud_id=compute_config.cloud_id,
        ).result.config

        # Disable head node scheduling.
        if default_compute_template.head_node_type.resources is None:
            default_compute_template.head_node_type.resources = {}
        default_compute_template.head_node_type.resources["CPU"] = 0

        # Ensure auto_select_worker_config is enabled.
        default_compute_template.auto_select_worker_config = True

        # Copy flags set at a cluster level over to the workspace.
        #
        # NOTE (shomilj): If there are more attributes we want to
        # persist from the provided compute config --> the compute
        # config used for deploying the service, we should copy them
        # over here.
        default_compute_template.flags = compute_config.flags

        return default_compute_template

    def get_cluster_env_build(self, build_id: str) -> Optional[ClusterEnvironmentBuild]:
        return self._external_api_client.get_cluster_environment_build(build_id).result

    def get_cluster_env_build_image_uri(
        self, cluster_env_build_id: str
    ) -> Optional[ImageURI]:
        try:
            build: ClusterEnvironmentBuild = self._external_api_client.get_cluster_environment_build(
                cluster_env_build_id
            ).result
            cluster_env = self._external_api_client.get_cluster_environment(
                build.cluster_environment_id
            ).result
            return ImageURI.from_cluster_env_build(cluster_env, build)
        except ExternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_default_build_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id
        result: ClusterEnvironmentBuild = self._external_api_client.get_default_cluster_environment_build(
            DEFAULT_PYTHON_VERSION, DEFAULT_RAY_VERSION,
        ).result
        return result.id

    def get_cluster_env_by_name(self, name: str) -> Optional[ClusterEnvironment]:
        resp = self._external_api_client.search_cluster_environments(
            ClusterEnvironmentsQuery(
                name=TextQuery(equals=name),  # pyright: ignore reportGeneralTypeIssues
                paging=PageQuery(count=1),
                include_anonymous=True,
            )
        )
        if resp.results:
            return resp.results[0]
        return None

    def list_cluster_env_builds(
        self, cluster_env_id: str,
    ) -> Generator[ClusterEnvironmentBuild, None, None]:
        paging_token = None
        while True:
            resp: ClusterenvironmentbuildListResponse = self._external_api_client.list_cluster_environment_builds(
                cluster_environment_id=cluster_env_id,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
                desc=True,
            )
            for build in resp.results:
                yield build

            if resp.metadata.next_paging_token is None:
                break
            paging_token = resp.metadata.next_paging_token

    def _wait_for_build_to_succeed(
        self, build_id: str, poll_interval_seconds=3, timeout_secs=3600,
    ):
        """Periodically check the status of the build operation until it completes.
        Raise a RuntimeError if the build fails or cancelled.
        Raise a TimeoutError if the build does not complete within the timeout.
        """
        elapsed_secs = 0
        while elapsed_secs < timeout_secs:
            build = self._external_api_client.get_cluster_environment_build(
                build_id
            ).result
            if build.status == ClusterEnvironmentBuildStatus.SUCCEEDED:
                self.logger.info("")
                return
            elif build.status == ClusterEnvironmentBuildStatus.FAILED:
                raise RuntimeError(f"Image build {build_id} failed.")
            elif build.status == ClusterEnvironmentBuildStatus.CANCELED:
                raise RuntimeError(f"Image build {build_id} unexpectedly cancelled.")

            elapsed_secs += poll_interval_seconds
            self.logger.info(
                f"Waiting for image build to complete. Elapsed time: {elapsed_secs} seconds.",
                end="\r",
            )
            self._sleep(poll_interval_seconds)
        raise TimeoutError(
            f"Timed out waiting for image build {build_id} to complete after {timeout_secs}s."
        )

    def _find_or_create_cluster_env(
        self, cluster_env_name: str, anonymous: bool
    ) -> ClusterEnvironment:
        existing_cluster_env = self.get_cluster_env_by_name(cluster_env_name)
        if existing_cluster_env is not None:
            return existing_cluster_env

        # this creates a cluster env only and it does not trigger a build to avoid the race condition
        # when creating a cluster environment and a build at the same time and then list builds for the cluster environment later.
        # ```
        #     cluster_env == self._external_api_client.create_cluster_environment(..., image_uri=image_uri)
        #     build == self._external_api_client.create_cluster_environment_build(..., cluster_env_id=cluster_env.id)
        # ```
        # The race condition can happen if another build for the same cluster envrionment is created between the two calls.
        cluster_environment = self._external_api_client.create_cluster_environment(
            CreateClusterEnvironment(name=cluster_env_name, anonymous=anonymous)
        ).result
        return cluster_environment

    def get_cluster_env_build_id_from_containerfile(
        self,
        cluster_env_name: str,
        containerfile: str,
        anonymous: bool = True,
        ray_version: Optional[str] = None,
    ) -> str:
        cluster_env = self._find_or_create_cluster_env(
            cluster_env_name, anonymous=anonymous
        )
        cluster_env_builds = self._external_api_client.list_cluster_environment_builds(
            cluster_environment_id=cluster_env.id, count=self.LIST_ENDPOINT_COUNT
        ).results
        for build in cluster_env_builds:
            if (
                build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and build.containerfile == containerfile
                # we don't need to check the ray_version because checking the containerfile is enough.
            ):
                return build.id

        try:
            build_op = self._external_api_client.create_cluster_environment_build(
                CreateClusterEnvironmentBuild(
                    cluster_environment_id=cluster_env.id,
                    containerfile=containerfile,
                    ray_version=ray_version,  # we don't use the latest version here if ray_version is Noneb/c the backend will try to parse the base image to decide the ray version.
                )
            ).result
        except ExternalApiException as e:
            if e.status == 400:
                raise RuntimeError(
                    "Invalid containerfile. Please check the syntax and try again.",
                    e.body,
                ) from None

        build_url = self.get_build_ui_url(
            cluster_env.id, build_op.cluster_environment_build_id
        )
        self.logger.info(f"Building image... See details: {build_url}.")
        self._wait_for_build_to_succeed(build_op.cluster_environment_build_id)
        self.logger.info("Image build succeeded.")

        return build_op.cluster_environment_build_id

    def get_cluster_env_build_id_from_image_uri(
        self,
        image_uri: ImageURI,
        registry_login_secret: Optional[str] = None,
        ray_version: Optional[str] = None,
    ) -> str:
        if image_uri.is_cluster_env_image():
            identifier = image_uri.to_cluster_env_identifier()
            try:
                build = self._external_api_client.find_cluster_environment_build_by_identifier(
                    identifier=identifier
                ).result
                if build.status == ClusterEnvironmentBuildStatus.SUCCEEDED:
                    return build.id
                else:
                    raise RuntimeError(
                        f"Legacy cluster environment build '{identifier}' is not a successful build."
                    )
            except ExternalApiException as e:
                if e.status == 404:
                    raise RuntimeError(
                        f"Legacy cluster environment '{identifier}' is not found."
                    )

        cluster_env_name = image_uri.to_cluster_env_name()
        cluster_env = self._find_or_create_cluster_env(cluster_env_name, anonymous=True)
        image_uri_str = str(image_uri)

        cluster_env_builds = self._external_api_client.list_cluster_environment_builds(
            cluster_environment_id=cluster_env.id, count=1
        ).results
        build = cluster_env_builds[0] if cluster_env_builds else None
        if (
            build is not None
            and build.docker_image_name == image_uri_str
            and build.registry_login_secret == registry_login_secret
            and build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
            and build.ray_version == ray_version
        ):
            return build.id

        # Still create a new build if the cluster env already exists but the build does not match the image_uri.
        result = self._external_api_client.create_cluster_environment_build(
            CreateClusterEnvironmentBuild(
                # For historical reasons, we have to use docker_image_name instead of image_uri; but it is just a URI to the image.
                cluster_environment_id=cluster_env.id,
                docker_image_name=image_uri_str,
                registry_login_secret=registry_login_secret,
                ray_version=ray_version if ray_version else LATEST_RAY_VERSION,
            )
        ).result

        assert result.completed
        return result.cluster_environment_build_id

    def send_workspace_notification(
        self, notification: WorkspaceNotification,
    ):
        if not self.inside_workspace():
            return

        try:
            r = requests.post(WORKSPACE_NOTIFICATION_ADDRESS, json=notification.dict())
            r.raise_for_status()
        except Exception:
            internal_logger.exception(
                "Failed to send workspace notification. "
                "This should not happen, so please contact Anyscale support."
            )

    def get_service(
        self, name: str, *, cloud: Optional[str], project: Optional[str]
    ) -> Optional[ServiceModel]:
        # TODO(edoakes): this endpoint is very slow and there's no reason we should need
        # to use this complex list endpoint just to fetch a service by name.
        paging_token = None
        cloud_id = self.get_cloud_id(cloud_name=cloud)
        project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
        service: Optional[ServiceModel] = None
        while True:
            resp = self._external_api_client.list_services(
                project_id=project_id,
                name=name,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
            )
            for result in resp.results:
                if result.name == name:
                    service = result
                    break

            paging_token = resp.metadata.next_paging_token
            if service is not None or paging_token is None:
                break

        return service

    def get_project(self, project_id: str) -> Optional[Project]:
        return self._internal_api_client.get_project_api_v2_projects_project_id_get(
            project_id
        ).result

    def get_job(
        self,
        *,
        name: Optional[str],
        job_id: Optional[str],
        cloud: Optional[str],
        project: Optional[str],
    ) -> Optional[ProductionJob]:
        if job_id is not None:
            try:
                return self._external_api_client.get_production_job(job_id).result
            except ExternalApiException as e:
                if e.status == 404:
                    return None
                raise e from None
        else:
            paging_token = None
            cloud_id = self.get_cloud_id(cloud_name=cloud)
            project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
            result: Optional[ProductionJob] = None
            while True:
                resp = self._external_api_client.list_production_jobs(
                    project_id=project_id,
                    name=name,
                    count=self.LIST_ENDPOINT_COUNT,
                    paging_token=paging_token,
                )
                for job in resp.results:
                    if (
                        job is not None
                        and job.name == name
                        and (result is None or job.created_at > result.created_at)
                    ):
                        result = job

                paging_token = resp.metadata.next_paging_token
                if paging_token is None:
                    break

            return result

    def get_job_runs(self, job_id: str) -> List[APIJobRun]:
        job_runs: List[APIJobRun] = search_entities(
            self._external_api_client.search_jobs,
            JobsQuery(
                ha_job_id=job_id,
                show_ray_client_runs_only=False,
                sort_by_clauses=[
                    SortByClauseJobsSortField(
                        sort_field=JobsSortField.CREATED_AT, sort_order=SortOrder.ASC,
                    )
                ],
                paging=PageQuery(),
            ),
        )
        return job_runs

    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollout_service(model).result
        return result

    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollback_service(
            service_id,
            rollback_service_model=RollbackServiceModel(
                max_surge_percent=max_surge_percent
            ),
        )
        return result

    def terminate_service(self, service_id: str) -> ServiceModel:
        result: ServiceModel = self._external_api_client.terminate_service(service_id)
        return result

    def submit_job(self, model: CreateInternalProductionJob) -> InternalProductionJob:
        job: InternalProductionJob = self._internal_api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            model,
        ).result
        return job

    def terminate_job(self, job_id: str):
        self._external_api_client.terminate_job(job_id)

    def archive_job(self, job_id: str):
        self._internal_api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post(
            job_id
        )

    def upload_local_dir_to_cloud_storage(
        self,
        local_dir: str,
        *,
        cloud_id: str,
        excludes: Optional[List[str]] = None,
        overwrite_existing_file: bool = OVERWRITE_EXISTING_CLOUD_STORAGE_FILES,
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        with zip_local_dir(local_dir, excludes=excludes) as (
            _,
            zip_file_bytes,
            content_hash,
        ):
            file_name = RUNTIME_ENV_PACKAGE_FORMAT.format(content_hash=content_hash)
            request = CloudDataBucketPresignedUploadRequest(
                file_type=CloudDataBucketFileType.RUNTIME_ENV_PACKAGES,
                file_name=file_name,
            )
            info: CloudDataBucketPresignedUploadInfo = self._internal_api_client.generate_cloud_data_bucket_presigned_upload_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_upload_url_post(
                cloud_id, request
            ).result

            # Skip the upload entirely if the zip file already exists.
            if not info.file_exists or overwrite_existing_file:
                internal_logger.debug(f"Uploading file '{file_name}' to cloud storage.")
                requests.put(info.upload_url, data=zip_file_bytes).raise_for_status()
            else:
                internal_logger.debug(
                    f"Skipping file upload for '{file_name}' because it already exists in cloud storage."
                )

        return info.file_uri

    def _fetch_log_chunks(self, job_run_id: str) -> Tuple[List[str], Any]:
        all_log_chunk_urls = []
        MAX_PAGE_SIZE = 1000
        next_page_token = None
        bearer_token = None
        while True:
            log_download_result = self._internal_api_client.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
                job_id=job_run_id,
                next_page_token=next_page_token,
                page_size=MAX_PAGE_SIZE,
            ).result

            if bearer_token is None:
                bearer_token = log_download_result.bearer_token

            all_log_chunk_urls.extend(
                [chunk.chunk_url for chunk in log_download_result.log_chunks]
            )

            next_page_token = log_download_result.next_page_token
            if next_page_token is None:
                break

        return all_log_chunk_urls, bearer_token

    def _read_log_lines(
        self,
        log_chunk_urls: List[str],
        head: bool,
        bearer_token: Any,
        max_lines: Optional[int],
    ) -> str:
        # TODO(mowen): Would be nice to return some placeholder here to allow all new
        # logs to be read from a particular point onwards.
        result_lines: List[str] = []
        step = 1 if head else -1
        line_count = 0
        for chunk_url in log_chunk_urls[::step]:
            log_lines = _download_log_from_s3_url_sync(
                chunk_url, bearer_token=bearer_token
            ).splitlines()

            if max_lines is not None:
                lines_to_add = min(len(log_lines), max_lines - line_count)
            else:
                lines_to_add = len(log_lines)
            line_count += lines_to_add

            if head:
                result_lines = result_lines + log_lines[:lines_to_add]
            else:
                result_lines = log_lines[-1 * lines_to_add :] + result_lines

            if line_count == max_lines:
                break

        return "\n".join(result_lines) + "\n"

    def logs_for_job_run(
        self, job_run_id: str, head: bool = False, max_lines: Optional[int] = None
    ) -> str:
        """Retrieves logs from the streaming job logs folder in S3/GCS"""

        all_log_chunk_urls, bearer_token = self._fetch_log_chunks(job_run_id)

        logs = self._read_log_lines(all_log_chunk_urls, head, bearer_token, max_lines)
        return logs

    def apply_schedule(self, model: CreateSchedule) -> DecoratedSchedule:
        try:
            return self._internal_api_client.create_or_update_job_api_v2_experimental_cron_jobs_put(
                model
            ).result
        except ApiException as e:
            if e.status == 400 and (
                "Please specify a valid cron expression" in e.body
                or "Please specify a valid tz database timezone" in e.body
            ):
                import json

                body_dict = json.loads(e.body)
                msg = body_dict["error"]["detail"]
                raise ValueError(msg) from None
            raise e from None

    def get_schedule(
        self,
        *,
        name: Optional[str],
        id: Optional[str],  # noqa: A002
        cloud: Optional[str],
        project: Optional[str],
    ) -> Optional[DecoratedSchedule]:
        if id is not None:
            try:
                return self._internal_api_client.get_cron_job_api_v2_experimental_cron_jobs_cron_job_id_get(
                    id
                ).result
            except ExternalApiException as e:
                if e.status == 404:
                    return None
                raise e from None
        else:
            paging_token = None
            cloud_id = self.get_cloud_id(cloud_name=cloud)
            project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
            result: Optional[DecoratedSchedule] = None
            while True:
                resp = self._internal_api_client.list_cron_jobs_api_v2_experimental_cron_jobs_get(
                    project_id=project_id,
                    name=name,
                    count=self.LIST_ENDPOINT_COUNT,
                    paging_token=paging_token,
                )
                for schedule in resp.results:
                    if schedule is not None and schedule.name == name:
                        result = schedule
                        break

                paging_token = resp.metadata.next_paging_token
                if paging_token is None:
                    break

            return result

    def set_schedule_state(self, id: str, is_paused: bool):  # noqa: A002
        self._internal_api_client.pause_cron_job_api_v2_experimental_cron_jobs_cron_job_id_pause_post(
            id, {"is_paused": is_paused}
        ).result

    def trigger_schedule(self, id: str):  # noqa: A002
        self._internal_api_client.trigger_cron_job_api_v2_experimental_cron_jobs_cron_job_id_trigger_post(
            id
        )



import click
import pytest

from anyscale.__init__ import _validate_headers
from anyscale.shared_anyscale_utils.headers import RequestHeaders


@pytest.mark.parametrize(
    ("headers", "raise_exception"),
    [
        ({RequestHeaders.SOURCE: "astronomy-sdk"}, False),
        ({RequestHeaders.SOURCE: 1}, True),
    ],
)
def test_validate_api_client_headers(headers, raise_exception) -> None:
    """
    Test headers are validated correctly.
    """
    if raise_exception:
        with pytest.raises(click.ClickException):
            _validate_headers(headers=headers)

    else:
        _validate_headers(headers=headers)



from typing import Optional, Union

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.compute_config._private.compute_config_sdk import PrivateComputeConfigSDK
from anyscale.compute_config.commands import (
    _ARCHIVE_ARG_DOCSTRINGS,
    _ARCHIVE_EXAMPLE,
    _CREATE_ARG_DOCSTRINGS,
    _CREATE_EXAMPLE,
    _GET_ARG_DOCSTRINGS,
    _GET_EXAMPLE,
    archive,
    create,
    get,
)
from anyscale.compute_config.models import (
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    WorkerNodeGroupConfig,
)


class ComputeConfigSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateComputeConfigSDK(
            client=client, logger=logger, timer=timer
        )

    @sdk_docs(
        doc_py_example=_CREATE_EXAMPLE, arg_docstrings=_CREATE_ARG_DOCSTRINGS,
    )
    def create(  # noqa: F811
        self, config: ComputeConfig, *, name: Optional[str],
    ) -> str:
        """Create a new version of a compute config.

        Returns the full name of the registered compute config, including the version.
        """
        full_name, _ = self._private_sdk.create_compute_config(config, name=name)
        return full_name

    @sdk_docs(
        doc_py_example=_GET_EXAMPLE, arg_docstrings=_GET_ARG_DOCSTRINGS,
    )
    def get(  # noqa: F811
        self, name: str, *, include_archived: bool = False, _id: Optional[str] = None,
    ) -> ComputeConfigVersion:
        """Get the compute config with the specified name.

        The name can contain an optional version tag, i.e., 'name:version'.
        If no version is provided, the latest one will be returned.
        """
        # NOTE(edoakes): I want to avoid exposing fetching by ID in the public API,
        # but it's needed for parity with the existing CLI. Therefore I am adding it
        # as a hidden private API that can be used like: (`name="", _id=id`).
        return self._private_sdk.get_compute_config(
            name=name or None, id=_id, include_archived=include_archived
        )

    @sdk_docs(
        doc_py_example=_ARCHIVE_EXAMPLE, arg_docstrings=_ARCHIVE_ARG_DOCSTRINGS,
    )
    def archive(self, name: str, *, _id: Optional[str] = None):  # noqa: F811
        """Archive a compute config and all of its versions.

        The name can contain an optional version, e.g., 'name:version'.
        If no version is provided, the latest one will be archived.

        Once a compute config is archived, its name will no longer be usable in the organization.
        """
        # NOTE(edoakes): I want to avoid exposing fetching by ID in the public API,
        # but it's needed for parity with the existing CLI. Therefore I am adding it
        # as a hidden private API that can be used like: (`name="", _id=id`).
        return self._private_sdk.archive_compute_config(name=name or None, id=_id,)



from typing import Optional

from anyscale._private.sdk import sdk_command
from anyscale.image._private.image_sdk import ImageSDK
from anyscale.image.models import ImageBuild


_IMAGE_SDK_SINGLETON_KEY = "image_sdk"

_BUILD_EXAMPLE = """
import anyscale

containerfile = '''
FROM anyscale/ray:2.21.0-py39
RUN pip install --no-cache-dir pandas
'''

image_uri: str = anyscale.image.build(containerfile, name="mycoolimage")
"""


@sdk_command(
    _IMAGE_SDK_SINGLETON_KEY,
    ImageSDK,
    doc_py_example=_BUILD_EXAMPLE,
    arg_docstrings={
        "name": "The name of the image.",
        "containerfile": "The content of the Containerfile.",
        "ray_version": "The version of Ray to use in the image",
    },
)
def build(
    containerfile: str, *, name: str, _sdk: ImageSDK, ray_version: Optional[str] = None
) -> str:
    """Build an image from a Containerfile.

    Returns the URI of the image.
    """
    build_id = _sdk.build_image_from_containerfile(
        name, containerfile, ray_version=ray_version
    )
    image_uri = _sdk.get_image_uri_from_build_id(build_id)
    if image_uri:
        return image_uri.image_uri
    raise RuntimeError(
        f"This is a bug! Failed to get image uri for build {build_id} that just created."
    )


_GET_EXAMPLE = """
import anyscale

image_status = anyscale.image.get(name="mycoolimage")
"""


@sdk_command(
    _IMAGE_SDK_SINGLETON_KEY,
    ImageSDK,
    doc_py_example=_GET_EXAMPLE,
    arg_docstrings={
        "name": (
            "Get the details of an image.\n\n"
            "The name can contain an optional version, e.g., 'name:version'. "
            "If no version is provided, the latest one will be used.\n\n"
        )
    },
)
def get(*, name: str, _sdk: ImageSDK) -> ImageBuild:
    """The name can contain an optional version tag, i.e., 'name:version'.

    If no version is provided, the latest one will be returned.
    """
    return _sdk.get(name)



import contextlib
import os
from typing import Any, Generator


@contextlib.contextmanager
def set_env(**environ: Any) -> Generator[None, None, None]:
    """
    Temporarily set the process environment variables.
    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)



from typing import Optional

from anyscale.anyscale_pydantic import BaseModel


class FineTuneConfig(BaseModel):
    base_model: str
    train_file: str
    valid_file: Optional[str]
    cloud_id: str
    suffix: Optional[str]
    version: Optional[str]
    instance_type: Optional[str]



import datetime
from datetime import timedelta
import os
import random
import shutil
import string
import tempfile
from typing import Any
from unittest.mock import call, patch

from anyscale.client.openapi_client.models import (
    LogDownloadConfig,
    LogDownloadRequest,
    LogDownloadResult,
    LogdownloadresultResponse,
    LogFileChunk,
    LogFilter,
)
from anyscale.controllers.logs_controller import LogsController


MOCK_COMBINED_LOGS = """
{"filename": "a", "message": "a"}
{"filename": "b", "message": "b"}
{"filename": "c", "message": "c"}
{"filename": "d", "message": "d"}
{"filename": "b", "message": "b"}
{"filename": "c", "message": "c"}
{"filename": "d", "message": "d"}
{"filename": "c", "message": "c"}
{"filename": "d", "message": "d"}
{"filename": "d", "message": "d"}
""".strip()

EXPECTED_UNCOMBINED_CONTENTS = {
    "a": "\n".join("a"),
    "b": "\n".join("bb"),
    "c": "\n".join("ccc"),
    "d": "\n".join("dddd"),
}


class AsyncBytesIterator:
    def __init__(self, data: bytes, n: int) -> None:
        self.data = data
        self.n = n
        self.read = 0

    def __aiter__(self):  # type: ignore
        return self

    async def __anext__(self) -> bytes:
        if not self.data[self.read :]:
            raise StopAsyncIteration
        _bytes = self.data[self.read : self.read + self.n]
        self.read += self.n
        return _bytes


class AsyncStreamReader:
    def __init__(self, _bytes: bytes) -> None:
        self._bytes = _bytes

    def iter_chunked(self, n: int) -> AsyncBytesIterator:
        return AsyncBytesIterator(self._bytes, n)


class AsyncMockResponse:
    def __init__(self, resp: str, status: int) -> None:
        self._bytes = bytes(resp, "utf-8")
        self.status = status
        self.content = AsyncStreamReader(self._bytes)

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        pass

    async def __aenter__(self):  # type: ignore
        return self

    async def text(self):  # type: ignore
        return self._bytes.decode("utf-8")


def test_download_log_files(mock_auth_api_client) -> None:  # type: ignore
    """Test that the download_log_files is able to download files correctly.
    Also verify handling of error response while downloading files.
    Also test handling of paginated response.
    """
    logs_controller = LogsController()
    data = ["".join(random.choices(string.ascii_lowercase, k=5)) for x in range(2)]
    data.append("error response")
    data.append(MOCK_COMBINED_LOGS)
    dirpath = tempfile.mkdtemp()
    try:
        logs_controller.api_client.get_log_files_api_v2_logs_get_log_files_post.side_effect = [
            LogdownloadresultResponse(
                result=LogDownloadResult(
                    log_chunks=[
                        LogFileChunk(
                            chunk_name="logs/test1.txt/chunk_one.log",
                            chunk_url="http://test.com",
                            size=len(data[0]),
                            cluster_id="cluster-id",
                            file_name="test1.txt",
                            node_type="head-node",
                            node_ip="ip",
                            instance_id="id",
                            session_id="session-id",
                        )
                    ],
                    next_page_token="foo",
                )
            ),
            LogdownloadresultResponse(
                result=LogDownloadResult(
                    log_chunks=[
                        LogFileChunk(
                            chunk_name="logs/events/test2.txt/chunk_one.log",
                            chunk_url="http://test.com",
                            size=len(data[1]),
                            cluster_id="cluster-id",
                            file_name="events/test2.txt",
                            node_type="head-node",
                            node_ip="ip",
                            instance_id="id",
                            session_id="session-id",
                        ),
                        LogFileChunk(
                            chunk_name="logs/test3.txt/chunk_one.log",
                            chunk_url="http://test.com",
                            size=len(data[2]),
                            cluster_id="cluster-id",
                            file_name="test3.txt",
                            node_type="head-node",
                            node_ip="ip",
                            instance_id="id",
                            session_id="session-id",
                        ),
                        LogFileChunk(
                            chunk_name="logs/combined-worker.log/chunk_one.log",
                            chunk_url="http://test.com",
                            size=len(data[3]),
                            cluster_id="cluster-id",
                            file_name="combined-worker.log",
                            node_type="head-node",
                            node_ip="ip",
                            instance_id="id",
                            session_id="session-id",
                        ),
                    ],
                    next_page_token=None,
                )
            ),
        ]

        with patch(
            "aiohttp.ClientSession.get",
            side_effect=[
                AsyncMockResponse(data[0], 200),
                AsyncMockResponse(data[1], 200),
                AsyncMockResponse(data[2], 500),
                AsyncMockResponse(data[3], 200),
            ],
        ):
            logs_controller.download_logs(
                filter=LogFilter(cluster_id="abc"),
                download_dir=dirpath,
                ttl_seconds=600,
                read_timeout=timedelta(5),
                parallelism=5,
                page_size=1000,
                timeout=timedelta(5),
                unpack=True,
            )
        with open(f"{dirpath}/logs/cluster-id/session-id/head-ip-id/test1.txt") as f:
            assert f.read().strip() == data[0].strip()
        with open(
            f"{dirpath}/logs/cluster-id/session-id/head-ip-id/events/test2.txt"
        ) as f:
            assert f.read().strip() == data[1].strip()
        for k, v in EXPECTED_UNCOMBINED_CONTENTS.items():
            with open(f"{dirpath}/logs/cluster-id/session-id/head-ip-id/{k}") as f:
                assert f.read().strip() == v

        assert not os.path.exists(
            f"{dirpath}/logs/cluster-id/session-id/head-ip-id/test3.txt"
        )
        assert (
            logs_controller.api_client.get_log_files_api_v2_logs_get_log_files_post.call_args_list
            == [
                call(
                    log_download_request=LogDownloadRequest(
                        filter=LogFilter(cluster_id="abc"),
                        config=LogDownloadConfig(
                            page_size=1000, ttl_seconds=600, use_bearer_token=True
                        ),
                    ),
                    _request_timeout=datetime.timedelta(days=5),
                ),
                call(
                    log_download_request=LogDownloadRequest(
                        filter=LogFilter(cluster_id="abc"),
                        config=LogDownloadConfig(
                            next_page_token="foo",
                            page_size=1000,
                            ttl_seconds=600,
                            use_bearer_token=True,
                        ),
                    ),
                    _request_timeout=datetime.timedelta(days=5),
                ),
            ]
        )

    finally:
        shutil.rmtree(dirpath)
